from core import Core
from event import Category, Event, Review, Video
from performance import Performance
from availability import TicketType, Concession, DespatchMethod, AvailDetail
from trolley import Trolley
from reservation import Reservation
from base import Customer, Seat, Card, Address, Commission, Currency
from bundle import Bundle
from order import Order

__all__ = (
    'Core', 'Category', 'Event', 'Review', 'Performance',
    'TicketType', 'Concession', 'DespatchMethod', 'AvailDetail',
    'Order', 'Trolley', 'Reservation', 'Customer', 'Commission',
    'Card', 'Address', 'Seat', 'Video', 'Bundle', 'Currency',
)



from settings import *

API_URL = 'https://dogbex.ingresso.co.uk/cgi-bin/xml_core.exe'
EXT_START_SESSION_URL = 'https://dogbex.ingresso.co.uk/cgi-bin/xml_start_session.exe'

TEST_USERNAME = 'whitelabel-testuser'
TEST_PASSWORD = ''

TEST_CREDIT_USERNAME = 'credit-testuser'
TEST_CREDIT_PASSWORD = 'testing123'



from base import InterfaceObject, Seat, Currency, Commission
from pyticketswitch.util import (
    to_int_or_return, to_float_or_none,
    to_float_summed, format_price_with_symbol
)
import performance as perf_objs
import event as event_objs
import availability


class Order(InterfaceObject):
    """Object that represents a TSW order.

    Information relating to a TSW order is accessible with this
    object, e.g. price information, seats, the event. Orders should
    be created with the Core.create_order method, the constructor is
    intended for internal use.
    """

    def __init__(
        self,
        order_id=None,
        core_order=None,
        core_currency=None,
        **settings
    ):
        self.order_id = order_id
        self._core_order = core_order
        self._core_currency = core_currency
        self._self_print_url = None
        self._self_print_relative_url = None
        self._requested_seats = False

        super(Order, self).__init__(**settings)

    @property
    def item_number(self):
        """Interger for identifying the order within the bundle."""
        return to_int_or_return(
            self._core_order.item_number
        )

    @property
    def self_print_url(self):
        if self._self_print_url:
            return self._self_print_url
        else:
            return None

    @self_print_url.setter
    def self_print_url(self, value):
        self._self_print_url = value

    @property
    def self_print_relative_url(self):
        if self._self_print_relative_url:
            return self._self_print_relative_url
        else:
            return None

    @self_print_relative_url.setter
    def self_print_relative_url(self, value):
        self._self_print_relative_url = value

    @property
    def event_desc(self):
        return self._core_order.event_desc

    @property
    def venue_desc(self):
        return self._core_order.venue_desc

    @property
    def total_combined_float(self):
        """Float value of the total combined price."""
        if self._core_order.total_combined:
            return to_float_or_none(
                self._core_order.total_combined
            )
        else:
            return to_float_summed(
                self._core_order.total_seatprice,
                self._core_order.total_surcharge
            )

    @property
    def total_combined(self):
        """Formatted string value of the total combined price with currency
        symbol.
        """
        return format_price_with_symbol(
            str(self.total_combined_float),
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def total_inc_despatch_float(self):
        """Float value of the total combined price including despatch."""
        total = self.total_combined_float
        if self.despatch_method and self.despatch_method.cost_float:
            total += self.despatch_method.cost_float

        return total

    @property
    def total_inc_despatch(self):
        """Formatted string value of the total combined price including
        despatch with currency symbol.
        """
        return format_price_with_symbol(
            str(self.total_inc_despatch_float),
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def currency(self):
        if self._core_currency:
            return Currency(
                core_currency=self._core_currency
            )

        return None

    @property
    def price_band_code(self):
        return self._core_order.price_band_code

    @property
    def average_price_per_ticket_float(self):
        return (
            self.total_combined_float /
            float(self._core_order.total_no_of_tickets)
        )

    @property
    def performance(self):
        """Performance object for this order."""
        if not hasattr(self, '_performance'):

            if self._core_order.performance:
                self._performance = perf_objs.Performance(
                    core_performance=self._core_order.performance,
                    **self._internal_settings()
                )

            else:
                self._performance = None

        return self._performance

    @property
    def event(self):
        """Event object for this order."""
        if not hasattr(self, '_event'):

            if self._core_order.event:

                self._event = event_objs.Event(
                    event_id=self._core_order.event.event_id,
                    core_event=self._core_order.event,
                    **self._internal_settings()
                )
            else:
                self._event = None

        return self._event

    @property
    def concessions(self):
        """List of Concession objects on this order."""
        if not hasattr(self, '_concessions'):

            self._concessions = []

            for discount in self._core_order.discounts:

                self._concessions.append(availability.Concession(
                    core_discount=discount,
                    core_currency=self._core_currency,
                    **self._internal_settings()
                ))

        return self._concessions

    @property
    def despatch_method(self):
        """DespatchMethod object for this order."""
        if not hasattr(self, '_despatch_method'):

            if self._core_order.despatch_method:

                self._despatch_method = availability.DespatchMethod(
                    core_despatch_method=self._core_order.despatch_method,
                    core_currency=self._core_currency,
                    **self._internal_settings()
                )

        return self._despatch_method

    @property
    def all_seats(self):
        """List of Seat objects on this order."""
        seats = []
        for con in self.concessions:
            seats = seats + con.seats

        return seats

    @property
    def all_seat_ids(self):
        """List of Seat Ids on this order."""
        return [s.seat_id for s in self.all_seats if s.seat_id]

    @property
    def ticket_type_desc(self):
        return self._core_order.ticket_type_desc

    @property
    def ticket_type_code(self):
        return self._core_order.ticket_type_code

    @property
    def backend_purchase_reference(self):
        """Supplier reference for this order."""
        return self._core_order.backend_purchase_reference

    @property
    def has_seat_with_restricted_view(self):
        """Boolean to indicate if any seats have restricted views."""
        restricted = False
        for con in self.concessions:
            if con.has_restricted_view:
                restricted = True

        return restricted

    @property
    def unique_seat_text(self):
        """List of unique seat text strings on this order."""
        seat_text = []
        for con in self.concessions:
            for text in con.unique_seat_text:
                if text not in seat_text:
                    seat_text.append(text)

        return seat_text

    @property
    def requested_seats(self):
        """If specific seats were requested, they will be listed here."""
        if self._requested_seats is False:
            if self._core_order.requested_seats:
                self._requested_seats = []

                for seat in self._core_order.requested_seats:
                    self._requested_seats.append(
                        Seat(core_seat=seat)
                    )
            else:
                self._requested_seats = None

        return self._requested_seats

    @property
    def requested_seat_ids(self):
        """List of requested Seat Ids on this order."""
        return [r.seat_id for r in self.requested_seats if r.seat_id]

    @property
    def seat_request_status(self):
        """Describes the status of the request for specific seats, i.e. were
        the specified seats successfully selected."""
        return self._core_order.seat_request_status

    @property
    def got_all_requested_seats(self):
        """Boolean indicating if the requested seats were successfully
        reserved.

        Returns True if all requested seats were reserved, False if some or
        none of the requested seats were reserved and None if no seats were
        requested.
        """
        if self.seat_request_status == 'not_requested':
            return None
        elif self.seat_request_status == 'got_all':
            return True
        return False

    @property
    def user_commission(self):
        """Returns Commission object representing the user commission for this
        order.
        """
        if self._core_order.user_commission:
            return Commission(
                core_commission=self._core_order.user_commission
            )
        return None

    @property
    def gross_commission(self):
        """Returns Commission object representing the gross commission for this
        order.
        """
        if self._core_order.gross_commission:
            return Commission(
                core_commission=self._core_order.gross_commission
            )
        return None



import core_objects as objects
from util import create_dict_from_xml_element
import api_exceptions as aex


def script_error(root):

    if root.tag == 'script_error':
        raise aex.APIException(
            call=root.tag,
            code=root.findtext('error_code'),
            description=root.findtext('error_desc')
        )
    else:
        return root


def error_check(root):
    fail_code = root.findtext('fail_code')

    if fail_code is not None:

        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    else:
        return root


def _text_dict(elem):
    return dict((e.tag, e.text) for e in list(elem))


def _parse_running_user(running_user_elem):
    return objects.RunningUser(**_text_dict(running_user_elem))


def start_session_resolve_user_result(root):
    root = error_check(root)

    ret_dict = {
        'crypto_block': root.findtext('crypto_block'),
        'country_code': root.findtext('country_code'),
        'country_desc': root.findtext('country_desc'),
    }

    running_user = root.find('running_user')

    if running_user is not None:
        ret_dict['running_user'] = _parse_running_user(running_user)

    return ret_dict


def style_map_result(root):
    root = error_check(root)

    ret_dict = {
        'mapping': []
    }

    for m in root.findall('mapping'):
        ret_dict['mapping'].append(_text_dict(m))

    return ret_dict


def _parse_geo_data(geo_elem):
    return objects.GeoData(**_text_dict(geo_elem))


def _parse_subclass(sub_elem):
    return objects.SubClass(**_text_dict(sub_elem))


def _parse_class(class_elem):
    subclasses = []
    for s in class_elem.findall('subclass'):
        subclasses.append(_parse_subclass(s))
        class_elem.remove(s)

    c_arg = _text_dict(class_elem)
    if subclasses:
        c_arg['subclasses'] = subclasses

    return objects.Class(**c_arg)


def _parse_event_media(event_media_elem):
    return objects.EventMedia(**_text_dict(event_media_elem))


def _parse_cost_range(cost_range_elem):
    add_arg = {}

    currency = cost_range_elem.find('range_currency')

    if currency is not None:
        add_arg['currency'] = _parse_currency(currency)
        cost_range_elem.remove(currency)

    no_singles = cost_range_elem.find('no_singles_cost_range')

    if no_singles is not None:
        add_arg['no_singles_cost_range'] = _parse_cost_range(no_singles)
        cost_range_elem.remove(no_singles)

    cr_arg = create_dict_from_xml_element(cost_range_elem)
    cr_arg.update(add_arg)

    return objects.CostRange(**cr_arg)


def _parse_review(review_elem):
    return objects.Review(**_text_dict(review_elem))


def _parse_video_iframe(video_iframe_elem):
    return objects.VideoIframe(**_text_dict(video_iframe_elem))


def _parse_custom_field(custom_field_elem):
    return objects.CustomField(**_text_dict(custom_field_elem))


def _parse_custom_filter(custom_filter_elem):
    return objects.CustomFilter(**_text_dict(custom_filter_elem))


def _parse_structured_info(structured_info_elem):

    ret_dict = {}

    for elem in structured_info_elem:

        si_args = _text_dict(elem)
        si_args['key'] = elem.tag

        ret_dict[elem.tag] = objects.StructuredInfoItem(
            **si_args
        )

    return ret_dict


def _parse_avail_detail(avail_detail_elem):

    ad_args = {}

    currency = avail_detail_elem.find('avail_currency')

    if currency is not None:
        ad_args['avail_currency'] = _parse_currency(currency)
        avail_detail_elem.remove(currency)

    ad_args.update(create_dict_from_xml_element(avail_detail_elem))
    return objects.AvailDetail(**ad_args)


def _parse_avail_details(avail_details_elem):
    """Parse the avail_details element. Does not make use of core TicketType
    or PriceBand since we only use a very limited subset of their data
    """

    ticket_types = []

    for tt in avail_details_elem.findall('ticket_type'):
        price_bands = []

        for pb in tt.findall('price_band'):
            avail_detail_list = []

            for ad in pb.findall('avail_detail'):
                avail_detail_list.append(_parse_avail_detail(ad))

            price_bands.append({
                'price_band_code': pb.findtext('price_band_code'),
                'price_band_desc': pb.findtext('price_band_desc'),
                'avail_details': avail_detail_list,
            })

        ticket_types.append({
            'ticket_type_code': tt.findtext('ticket_type_code'),
            'ticket_type_desc': tt.findtext('ticket_type_desc'),
            'price_bands': price_bands,
        })

    return {'ticket_types': ticket_types}


def _parse_event(event_elem):

    objs = {}

    classes = []
    for c in event_elem.findall('class'):

        classes.append(_parse_class(c))
        event_elem.remove(c)

    objs['classes'] = classes

    event_medias = []
    for event_media in event_elem.findall('event_media'):
        event_medias.append(_parse_event_media(event_media))
        event_elem.remove(event_media)

    objs['event_medias'] = event_medias

    reviews = event_elem.find('reviews')

    if reviews is not None:
        review_list = []

        for review in reviews.findall('review'):
            review_list.append(_parse_review(review))

        event_elem.remove(reviews)
        objs['reviews'] = review_list

    geo_elem = event_elem.find('geo_data')

    if geo_elem is not None:
        objs['geo_data'] = _parse_geo_data(geo_elem)
        event_elem.remove(geo_elem)

    cost_range = event_elem.find('cost_range')

    if cost_range is not None:
        objs['cost_range'] = _parse_cost_range(cost_range)
        event_elem.remove(cost_range)

    custom_fields = []

    for custom_field in event_elem.findall('custom_field'):
        custom_fields.append(_parse_custom_field(custom_field))
        event_elem.remove(custom_field)

    objs['custom_fields'] = custom_fields

    custom_filters = []

    for custom_filter in event_elem.findall('custom_filter'):
        custom_filters.append(_parse_custom_filter(custom_filter))
        event_elem.remove(custom_filter)

    objs['custom_filters'] = custom_filters

    video_iframe = event_elem.find('video_iframe')

    if video_iframe is not None:
        objs['video_iframe'] = _parse_video_iframe(video_iframe)
        event_elem.remove(video_iframe)

    structured_info = event_elem.find('structured_info')

    if structured_info is not None:
        objs['structured_info'] = _parse_structured_info(structured_info)
        event_elem.remove(structured_info)

    avail_details = event_elem.find('avail_details')

    if avail_details is not None:
        objs['avail_details'] = _parse_avail_details(avail_details)
        event_elem.remove(avail_details)

    e_arg = create_dict_from_xml_element(event_elem)
    e_arg.update(objs)

    return objects.Event(**e_arg)


def event_search_result(root):
    root = error_check(root)

    ret_dict = {
        'crypto_block': root.findtext('crypto_block'),
        'event': []
    }

    for e in root.findall('event'):
        ret_dict['event'].append(_parse_event(e))

    return ret_dict


def extra_info_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code in (
        '102', '103',
    ):
        raise aex.InvalidToken(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    event = _parse_event(root)

    return event


def _parse_performance(perf_elem):

    objs = {}

    cost_range = perf_elem.find('cost_range')

    if cost_range is not None:
        objs['cost_range'] = _parse_cost_range(cost_range)
        perf_elem.remove(cost_range)

    p_arg = create_dict_from_xml_element(perf_elem)
    p_arg.update(objs)

    return objects.Performance(**p_arg)


def date_time_options_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code in (
        '202', '203'
    ):
        raise aex.InvalidToken(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    ret_dict = {
        'crypto_block': root.findtext('crypto_block'),
        'need_departure_date': root.findtext('need_departure_date'),
        'has_perf_names': root.findtext('has_perf_names'),
    }

    using_perf_list = root.find('using_perf_list')
    using_usage_date = root.find('using_usage_date')

    if using_perf_list is not None:
        perf_list = []
        for p in using_perf_list.findall('performance'):
            perf_list.append(_parse_performance(p))

        ret_dict['using_perf_list'] = {'performances': perf_list}

    elif using_usage_date is not None:
        ret_dict['using_usage_date'] = create_dict_from_xml_element(
            using_usage_date
        )

    return ret_dict


def _parse_month(month_elem):
    return objects.Month(**_text_dict(month_elem))


def month_options_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code in (
        '202', '203'
    ):
        raise aex.InvalidToken(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    ret_dict = {
        'need_departure_date': root.findtext('need_departure_date'),
        'has_perf_names': root.findtext('has_perf_names'),
    }

    using_perf_list = root.find('using_perf_list')

    if using_perf_list is not None:
        month_list = []
        for m in using_perf_list.findall('month'):
            month_list.append(_parse_month(m))

        ret_dict['months'] = month_list

    return ret_dict


def _parse_country(country_elem):
    return objects.Country(**_text_dict(country_elem))


def _parse_currency(curr_elem):
    return objects.Currency(**_text_dict(curr_elem))


def _parse_despatch_method(desp_elem):

    d_arg = _text_dict(desp_elem)

    permitted_countries = desp_elem.find('permitted_countries')

    if permitted_countries is not None:
        country = []
        for c in permitted_countries.findall('country'):
            country.append(_parse_country(c))

        d_arg['permitted_countries'] = {'country': country}

    return objects.DespatchMethod(**d_arg)


def _parse_seats(seats_elem):

    seats = []

    for s in seats_elem.findall('id_details'):
        seats.append(objects.Seat(**_text_dict(s)))

    return seats


def _parse_seat_block(seat_block_elem):

    sb_arg = {}

    seats = []

    for s in seat_block_elem.findall('id_details'):
        seats.append(objects.Seat(**_text_dict(s)))
        seat_block_elem.remove(s)

    sb_arg = _text_dict(seat_block_elem)

    if seats:
        sb_arg['seats'] = seats

    return objects.SeatBlock(**sb_arg)


def _parse_free_seat_blocks(free_seat_blocks_elem):

    seat_blocks = []

    for sb in free_seat_blocks_elem.findall('seat_block'):
        seat_blocks.append(_parse_seat_block(sb))

    return seat_blocks


def _parse_commission(commission_elem):

    c_args = {}

    currency = commission_elem.find('commission_currency')

    if currency is not None:
        c_args['commission_currency'] = _parse_currency(currency)
        commission_elem.remove(currency)

    c_args.update(_text_dict(commission_elem))

    return objects.Commission(**c_args)


def _parse_price_band(price_elem):

    objs = {}

    possible_discounts = price_elem.find('possible_discounts')

    if possible_discounts is not None:
        discounts_list = []
        for d in possible_discounts.findall('discount'):
            discounts_list.append(_parse_discount(d))

        objs['possible_discounts'] = discounts_list
        price_elem.remove(possible_discounts)

    example_seats = price_elem.find('example_seats')

    if example_seats is not None:
        objs['example_seats'] = _parse_seats(example_seats)
        price_elem.remove(example_seats)

    free_seat_blocks = price_elem.find('free_seat_blocks')

    if free_seat_blocks is not None:
        objs['free_seat_blocks'] = _parse_free_seat_blocks(
            free_seat_blocks
        )
        price_elem.remove(free_seat_blocks)

    user_commission = price_elem.find('user_commission')
    if user_commission is not None:
        objs['user_commission'] = _parse_commission(
            user_commission
        )
        price_elem.remove(user_commission)

    gross_commission = price_elem.find('gross_commission')
    if gross_commission is not None:
        objs['gross_commission'] = _parse_commission(
            gross_commission
        )
        price_elem.remove(gross_commission)

    p_arg = _text_dict(price_elem)
    p_arg.update(objs)

    return objects.PriceBand(**p_arg)


def _parse_ticket_type(tt_elem):

    price_bands = []

    for p in tt_elem.findall('price_band'):
        price_bands.append(_parse_price_band(p))
        tt_elem.remove(p)

    t_arg = _text_dict(tt_elem)

    t_arg['price_bands'] = price_bands

    return objects.TicketType(**t_arg)


def availability_options_result(root):
    root = error_check(root)

    backend_call_failed = root.find('backend_call_failed')

    if backend_call_failed:
        raise aex.BackendCallFailure(call=root.tag)

    ret_dict = {
        'crypto_block': root.findtext('crypto_block')
    }

    quantity_options = root.find('quantity_options')

    if quantity_options is not None:
        ret_dict['quantity_options'] = create_dict_from_xml_element(
            quantity_options
        )

    availability = root.find('availability')

    if availability is not None:

        ticket_types = []
        for t in availability.findall('ticket_type'):

            ticket_types.append(_parse_ticket_type(t))

        ret_dict['ticket_type'] = ticket_types

    despatch_options = root.find('despatch_options')

    if despatch_options is not None:
        despatch_method = []
        for d in despatch_options.findall('despatch_method'):

            despatch_method.append(_parse_despatch_method(d))

        ret_dict['despatch_options'] = {'despatch_method': despatch_method}

    currency = root.find('currency')

    if currency is not None:
        ret_dict['currency'] = _parse_currency(currency)

    event = root.find('event')

    if event is not None:
        ret_dict['event'] = _parse_event(event)

    performance = root.find('performance')

    if performance is not None:
        ret_dict['performance'] = _parse_performance(performance)

    return ret_dict


def despatch_options_result(root):
    root = error_check(root)

    ret_dict = {}

    despatch_options = root.find('despatch_options')

    if despatch_options is not None:
        despatch_method = []
        for d in despatch_options.findall('despatch_method'):

            despatch_method.append(_parse_despatch_method(d))

        ret_dict['despatch_options'] = {'despatch_method': despatch_method}

    currency = root.find('currency')

    if currency is not None:
        ret_dict['currency'] = _parse_currency(currency)

    event = root.find('event')

    if event is not None:
        ret_dict['event'] = _parse_event(event)

    performance = root.find('performance')

    if performance is not None:
        ret_dict['performance'] = _parse_performance(performance)

    return ret_dict


def _parse_discount(discount_elem):

    d_arg = _text_dict(discount_elem)

    seats = discount_elem.find('seats')

    if seats is not None:
        d_arg['seats'] = _parse_seats(seats)

    user_commission = discount_elem.find('user_commission')
    if user_commission is not None:
        d_arg['user_commission'] = _parse_commission(
            user_commission
        )

    gross_commission = discount_elem.find('gross_commission')
    if gross_commission is not None:
        d_arg['gross_commission'] = _parse_commission(
            gross_commission
        )

    return objects.Discount(**d_arg)


def discount_options_result(root):
    root = error_check(root)

    backend_call_failed = root.find('backend_call_failed')

    if backend_call_failed:
        raise aex.BackendCallFailure(call=root.tag)

    ret_dict = {
        'crypto_block': root.findtext('crypto_block')
    }

    blanket_discount_only = root.findtext('blanket_discount_only')

    if blanket_discount_only is not None:
        ret_dict['blanket_discount_only'] = blanket_discount_only

    currency = root.find('currency')

    if currency is not None:
        ret_dict['currency'] = _parse_currency(currency)

    discount_limit = root.findtext('discount_limit')

    if discount_limit is not None:
        ret_dict['discount_limit'] = discount_limit

    discounts_list = []
    for td in root.findall('discounts'):
        discount_objs = []
        for d in td.findall('discount'):
            discount_objs.append(_parse_discount(d))

        discounts_list.append(discount_objs)

    ret_dict['discounts'] = discounts_list

    return ret_dict


def _parse_order(order_elem):

    objs = {}

    despatch_method = order_elem.find('despatch_method')

    if despatch_method is not None:
        objs['despatch_method'] = _parse_despatch_method(despatch_method)
        order_elem.remove(despatch_method)

    discounts_list = []
    for discount in order_elem.findall('discount'):
        discounts_list.append(_parse_discount(discount))
        order_elem.remove(discount)

    objs['discounts'] = discounts_list

    performance = order_elem.find('performance')

    if performance is not None:
        objs['performance'] = _parse_performance(performance)
        order_elem.remove(performance)

    event = order_elem.find('event')

    if event is not None:
        objs['event'] = _parse_event(event)
        order_elem.remove(event)

    requested_seats = order_elem.find('requested_seats')

    if requested_seats is not None:
        objs['requested_seats'] = _parse_seats(requested_seats)
        order_elem.remove(requested_seats)

    user_commission = order_elem.find('user_commission')
    if user_commission is not None:
        objs['user_commission'] = _parse_commission(
            user_commission
        )
        order_elem.remove(user_commission)

    gross_commission = order_elem.find('gross_commission')
    if gross_commission is not None:
        objs['gross_commission'] = _parse_commission(
            gross_commission
        )
        order_elem.remove(gross_commission)

    o_arg = _text_dict(order_elem)
    o_arg.update(objs)

    return objects.Order(**o_arg)


def create_order_result(root):
    root = error_check(root)

    ret_dict = {
        'crypto_block': root.findtext('crypto_block')
    }

    order = root.find('order')

    if order is not None:
        ret_dict['order'] = _parse_order(order)

    ret_dict['order_token'] = root.findtext('order_token')

    currency = root.find('currency')

    if currency is not None:
        ret_dict['currency'] = _parse_currency(currency)

    return ret_dict


def create_order_and_reserve_result(root):
    return make_reservation_result(root)


def trolley_add_order_result(root):
    root = error_check(root)

    ret_dict = {
        'crypto_block': root.findtext('crypto_block'),
        'add_possible': root.findtext('add_possible'),
        'trolley_token': root.findtext('trolley_token'),
        'trolley_order_count': root.findtext('trolley_order_count'),
        'added_item_number': root.findtext('added_item_number'),
    }

    if ret_dict['add_possible'] == 'no':
        errors = []

        if root.findtext('trolley_bad_bundle') == 'yes':
            errors.append(
                aex.TrolleyAddBadBundle(
                    root.findtext('trolley_bad_bundle_max_size')
                )
            )

        if root.findtext('trolley_bad_combo') == 'yes':
            errors.append(
                aex.TrolleyAddBadCombo(
                    system=root.findtext('trolley_bad_combo_system'),
                    system_desc=root.findtext('trolley_bad_combo_system_desc')
                )
            )

        if root.findtext('trolley_bad_card_types') == 'yes':
            errors.append(aex.TrolleyAddBadCardTypes())

        if root.findtext('trolley_bad_countries') == 'yes':
            errors.append(aex.TrolleyAddBadCountries())

        if root.findtext('trolley_bad_currency_mix') == 'yes':

            curr_elem = root.find('trolley_bad_currency')
            if curr_elem:
                currency = _parse_currency(curr_elem)
            else:
                currency = None

            errors.append(
                aex.TrolleyAddBadCurrencyMix(
                    system=root.findtext(
                        'trolley_bad_currency_system'
                    ),
                    system_desc=root.findtext(
                        'trolley_bad_currency_system_desc'
                    ),
                    currency=currency,
                    currency_name=root.findtext(
                        'trolley_bad_currency_name'
                    )
                )
            )

        if root.findtext('trolley_bad_depart') == 'yes':
            errors.append(aex.TrolleyAddBadDepart())

        if root.findtext('trolley_bad_send') == 'yes':
            errors.append(aex.TrolleyAddBadSend())

        if errors:
            raise aex.TrolleyAddErrors(errors)

    trolley = root.find('trolley')

    if trolley is not None:
        ret_dict['trolley'] = _parse_trolley(trolley)

    return ret_dict


def _parse_trolley(trolley_elem):

    objs = {}

    purchase_result = trolley_elem.find('purchase_result')

    if purchase_result is not None:
        objs['purchase_result'] = _parse_purchase_result(purchase_result)
        trolley_elem.remove(purchase_result)

    bundles = []

    for bundle in trolley_elem.findall('bundle'):

        bundles.append(_parse_bundle(bundle))
        trolley_elem.remove(bundle)

    objs['bundles'] = bundles

    t_arg = _text_dict(trolley_elem)
    t_arg.update(objs)

    return objects.Trolley(**t_arg)


def _parse_purchase_result(purchase_result_elem):
    return objects.PurchaseResult(**_text_dict(purchase_result_elem))


def _parse_bundle(bundle_elem):

    objs = {}

    purchase_result = bundle_elem.find('purchase_result')

    if purchase_result is not None:
        objs['purchase_result'] = _parse_purchase_result(purchase_result)
        bundle_elem.remove(purchase_result)

    currency = bundle_elem.find('currency')

    if currency is not None:
        objs['currency'] = _parse_currency(currency)
        bundle_elem.remove(currency)

    orders = []

    for order in bundle_elem.findall('order'):
        orders.append(_parse_order(order))
        bundle_elem.remove(order)

    objs['orders'] = orders

    b_arg = _text_dict(bundle_elem)
    b_arg.update(objs)

    return objects.Bundle(**b_arg)


def trolley_describe_result(root):
    root = error_check(root)

    trolley = root.find('trolley')

    ret_dict = {
        'trolley': _parse_trolley(trolley)
    }

    return ret_dict


def trolley_remove_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code == '803':
        raise aex.TrolleyPurchased(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '804':
        raise aex.TrolleyReserved(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    ret_dict = {
        'crypto_block': root.findtext('crypto_block'),
        'trolley_token': root.findtext('trolley_token'),
        'trolley_order_count': root.findtext('trolley_order_count'),
    }

    trolley = root.find('trolley')

    if trolley is not None:
        ret_dict['trolley'] = _parse_trolley(trolley)

    return ret_dict


def make_reservation_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code == '903':
        raise aex.TrolleyPurchased(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '904':
        raise aex.TrolleyReserved(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    objs = {}

    trolley = root.find('trolley')

    if trolley is not None:
        objs['trolley'] = _parse_trolley(trolley)
        root.remove(trolley)

    failed_orders = root.find('failed_orders')

    if failed_orders is not None:
        failed_list = []
        for order in failed_orders.findall('order'):
            failed_list.append(_parse_order(order))
            failed_orders.remove(order)

        objs['failed_orders'] = failed_list
        root.remove(failed_orders)

    ret_dict = create_dict_from_xml_element(root)
    ret_dict.update(objs)

    avail_test = [
        x for x in ret_dict.keys() if x not in (
            'subdomain_user_is_bad', 'currency'
        )
    ]

    if not avail_test:
        raise aex.NoAvailability(
            call=root.tag
        )

    return ret_dict


def get_reservation_link_result(root):
    root = error_check(root)

    return create_dict_from_xml_element(root)


def release_reservation_result(root):
    root = error_check(root)

    return create_dict_from_xml_element(root)


def purchase_reservation_part_one_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code == '1301':
        raise aex.ReservationExpired(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1306':
        raise aex.InvalidCountryForDespatch(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1307':
        raise aex.InvalidEmailAddress(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1308':
        raise aex.IncompleteCustomerDetails(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1309':
        raise aex.NoCardNumberProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1310':
        raise aex.UnknownCardType(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1311':
        raise aex.InvalidCardType(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1312':
        raise aex.InvalidCardNumber(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1313':
        raise aex.NoExpiryDateProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1314':
        raise aex.InvalidExpiryDate(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1315':
        raise aex.NoCV2Provided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1316':
        raise aex.InvalidCV2(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1317':
        raise aex.NoIssueNumberProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1319':
        raise aex.InvalidIssueNumber(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1321':
        raise aex.IncompleteAltBillingAddress(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1322':
        raise aex.NoStartDateProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1323':
        raise aex.InvalidStartDate(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    return create_dict_from_xml_element(root)


def _parse_customer(customer_elem):
    return objects.Customer(**_text_dict(customer_elem))


def _parse_self_print_html_page(self_print_html_page_elem):
    return objects.SelfPrintHTMLPage(**_text_dict(self_print_html_page_elem))


def purchase_reservation_part_two_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code == '1404':
        raise aex.ReservationExpired(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    failed_cv_two = root.findtext('failed_cv_two')
    failed_avs = root.findtext('failed_avs')
    failed_3d_secure = root.findtext('failed_3d_secure')
    purchase_fail_code = root.findtext('purchase_fail_code')

    if purchase_fail_code == '1':
        raise aex.FraudTriggered(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code in ('2', '3'):
        raise aex.CardAuthorisationFailed(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '4':
        raise aex.ReservationAlreadyPurchased(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '5':
        raise aex.PurchasePreviouslyAttempted(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '6':
        raise aex.PurchaseRefused(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '7':
        raise aex.PurchaseFailed(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code:
        raise aex.PurchaseException(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )

    objs = {}

    self_print_html_pages = []

    for self_print_html_page in root.findall('self_print_html_page'):

        self_print_html_pages.append(
            _parse_self_print_html_page(self_print_html_page)
        )
        root.remove(self_print_html_page)

    objs['self_print_html_pages'] = self_print_html_pages

    trolley = root.find('trolley')

    if trolley is not None:
        objs['trolley'] = _parse_trolley(trolley)
        root.remove(trolley)

    customer = root.find('customer')

    if customer is not None:
        objs['customer'] = _parse_customer(customer)
        root.remove(customer)

    ret_dict = create_dict_from_xml_element(root)
    ret_dict.update(objs)

    return ret_dict


def purchase_reservation(root):
    fail_code = root.findtext('fail_code')

    if fail_code == '1101':
        raise aex.ReservationExpired(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1106':
        raise aex.InvalidCountryForDespatch(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1107':
        raise aex.InvalidEmailAddress(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1108':
        raise aex.IncompleteCustomerDetails(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1109':
        raise aex.NoCardNumberProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1110':
        raise aex.UnknownCardType(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1111':
        raise aex.InvalidCardType(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1112':
        raise aex.InvalidCardNumber(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1113':
        raise aex.NoExpiryDateProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1114':
        raise aex.InvalidExpiryDate(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1115':
        raise aex.NoCV2Provided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1116':
        raise aex.InvalidCV2(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1117':
        raise aex.NoIssueNumberProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1119':
        raise aex.InvalidIssueNumber(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1121':
        raise aex.IncompleteAltBillingAddress(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1122':
        raise aex.NoStartDateProvided(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code == '1123':
        raise aex.InvalidStartDate(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    failed_cv_two = root.findtext('failed_cv_two')
    failed_avs = root.findtext('failed_avs')
    failed_3d_secure = root.findtext('failed_3d_secure')
    purchase_fail_code = root.findtext('purchase_fail_code')

    if purchase_fail_code == '1':
        raise aex.FraudTriggered(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code in ('2', '3'):
        raise aex.CardAuthorisationFailed(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '4':
        raise aex.ReservationAlreadyPurchased(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '5':
        raise aex.PurchasePreviouslyAttempted(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '6':
        raise aex.PurchaseRefused(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code == '7':
        raise aex.PurchaseFailed(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )
    elif purchase_fail_code:
        raise aex.PurchaseException(
            call=root.tag,
            code=purchase_fail_code,
            description=root.findtext('purchase_fail_desc'),
            failed_cv_two=failed_cv_two,
            failed_avs=failed_avs,
            failed_3d_secure=failed_3d_secure,
        )

    objs = {}

    self_print_html_pages = []

    for self_print_html_page in root.findall('self_print_html_page'):

        self_print_html_pages.append(
            _parse_self_print_html_page(self_print_html_page)
        )
        root.remove(self_print_html_page)

    objs['self_print_html_pages'] = self_print_html_pages

    trolley = root.find('trolley')

    if trolley is not None:
        objs['trolley'] = _parse_trolley(trolley)
        root.remove(trolley)

    customer = root.find('customer')

    if customer is not None:
        objs['customer'] = _parse_customer(customer)
        root.remove(customer)

    ret_dict = create_dict_from_xml_element(root)
    ret_dict.update(objs)

    return ret_dict


def _parse_sale_page(sale_page_elem):
    return objects.SalePage(**_text_dict(sale_page_elem))


def transaction_info_result(root):
    fail_code = root.findtext('fail_code')

    if fail_code in ('1201', '1202'):
        raise aex.InvalidToken(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )
    elif fail_code:
        raise aex.APIException(
            call=root.tag,
            code=fail_code,
            description=root.findtext('fail_desc')
        )

    ret_dict = {
        'transaction_status': root.findtext('transaction_status'),
        'minutes_left_on_reserve': root.findtext('minutes_left_on_reserve'),
        'language_list': root.findtext('language_list'),
        'remote_site': root.findtext('remote_site'),
    }

    objs = {}

    self_print_html_pages = []

    for self_print_html_page in root.findall('self_print_html_page'):

        self_print_html_pages.append(
            _parse_self_print_html_page(self_print_html_page)
        )
        root.remove(self_print_html_page)

    objs['self_print_html_pages'] = self_print_html_pages

    trolley = root.find('trolley')

    if trolley is not None:
        objs['trolley'] = _parse_trolley(trolley)
        root.remove(trolley)

    customer = root.find('customer')

    if customer is not None:
        objs['customer'] = _parse_customer(customer)
        root.remove(customer)

    external_sale_page = root.find('external_sale_page')

    if external_sale_page is not None:
        if len(list(external_sale_page)) > 0:
            objs['sale_page'] = _parse_sale_page(external_sale_page)
            root.remove(external_sale_page)

    ret_dict.update(objs)

    return ret_dict


def save_external_sale_page_result(root):
    root = error_check(root)

    return create_dict_from_xml_element(root)



from operator import attrgetter
import datetime

from base import InterfaceObject, CostRangeMixin
from pyticketswitch.util import (
    yyyymmdd_to_date, hhmmss_to_time,
    date_to_yyyymmdd_or_none, time_to_hhmmss,
    date_to_yyyymmdd, resolve_boolean, to_int_or_none
)
import event as event_objs
import availability


class Performance(InterfaceObject, CostRangeMixin):
    """Object that represents a TSW performance.

    Operations relating to a TSW performance are accessible through this
    object, e.g. getting available ticket types. Can be constructed with
    the Id of the Performance (other constructor arguments are for
    internal use).

    Inherits from the CostRangeMixin class, so these functions are
    available.

    Args:
        perf_id (string): Id of the Performance
        settings (kwargs): See Core constructor.
    """

    @staticmethod
    def _get_perf_id(
        event_id, perf_token=None, usage_date=None, departure_date=None
    ):

        perf_id = '{0}{1}'.format(
            len(event_id), event_id,
        )

        if departure_date:
            perf_id = '{0}d{1}'.format(
                perf_id, date_to_yyyymmdd(departure_date)
            )

        if perf_token:
            perf_id = '{0}p{1}'.format(
                perf_id, perf_token
            )

        elif usage_date:
            perf_id = '{0}u{1}'.format(
                perf_id, date_to_yyyymmdd(usage_date)
            )

        return perf_id

    def __init__(
        self,
        perf_id=None,
        core_performance=None,
        **settings
    ):
        self.perf_id = perf_id
        self._set_perf_id_attrs(perf_id)
        self._core_performance = core_performance

        self._ticket_types = None
        self._despatch_methods = None
        self._valid_ticket_quantities = None
        self._event = None

        super(Performance, self).__init__(**settings)

        if core_performance is not None:
            self._set_datetime(
                date_yyyymmdd=core_performance.date_yyyymmdd,
                time_hhmmss=core_performance.time_hhmmss
            )
        elif self.usage_date is not None:
            self.date = self.usage_date

        else:
            self._set_datetime(None)

        if self._event_id:
            self._event = event_objs.Event(
                event_id=self._event_id,
                **self._internal_settings()
            )

    @classmethod
    def from_event_and_perf_token(
        cls,
        event,
        perf_token,
        core_performance,
        departure_date=None,
        **settings
    ):

        return cls(
            perf_id=cls._get_perf_id(
                event_id=event.event_id,
                perf_token=perf_token,
                departure_date=departure_date,
            ),
            core_performance=core_performance,
            **settings
        )

    @classmethod
    def from_event_and_usage_date(
        cls,
        event,
        usage_date,
        departure_date=None,
        **settings
    ):

        return cls(
            perf_id=cls._get_perf_id(
                event_id=event.event_id,
                usage_date=usage_date,
                departure_date=departure_date,
            ),
            **settings
        )

    @classmethod
    def from_event_only(
        cls, event, **settings
    ):

        return cls(
            perf_id=cls._get_perf_id(
                event_id=event.event_id,
            ),
            **settings
        )

    def _set_perf_id_attrs(self, perf_id):

        self._event_id = None
        self.departure_date = None
        self._perf_token = None
        self.usage_date = None

        if perf_id:

            event_len = int(perf_id[0])

            self._event_id = perf_id[1:event_len + 1]

            remaining = perf_id[event_len + 1:]

            if remaining and remaining[0] == 'd':
                self.departure_date = yyyymmdd_to_date(
                    remaining[1:9]  # yyyymmdd
                )
                remaining = remaining[9:]

            if remaining and remaining[0] == 'p':
                self._perf_token = remaining[1:]

            elif remaining and remaining[0] == 'u':
                self.usage_date = yyyymmdd_to_date(
                    remaining[1:9]  # yyyymmdd
                )

    def _set_datetime(self, date_yyyymmdd, time_hhmmss=None):

        if date_yyyymmdd:
            self.date = yyyymmdd_to_date(date_yyyymmdd)

        else:
            self.date = None

        if time_hhmmss:
            self.time = hhmmss_to_time(time_hhmmss)

        else:
            self.time = None

    def _get_cache_key(self):
        return self.perf_id

    def _get_core_performance_attr(self, attr):
        return getattr(self._core_performance, attr, None)

    # Overridden function for CostRangeMixin
    def _get_core_cost_range(self):
        return self._get_core_performance_attr('cost_range')

    @property
    def date_desc(self):
        return self._get_core_performance_attr('date_desc')

    @property
    def time_desc(self):
        return self._get_core_performance_attr('time_desc')

    @property
    def date_slug(self):
        return date_to_yyyymmdd(self.date)

    @property
    def time_slug(self):
        return time_to_hhmmss(self.time)

    @property
    def day_name(self):
        if self.date is not None:
            if self.delta_from_today == 0:
                return 'Today'
            elif self.delta_from_today == 1:
                return 'Tomorrow'
            else:
                return self.date.strftime('%A')

    @property
    def delta_from_today(self):
        """Integer number of days from today."""
        if self.date is not None:
            today = datetime.date.today()

            return (self.date - today).days

    @property
    def required_info(self):
        """Required info text.

        Returns a string of the required_info or None if there isn't
        any for this performance. Must be displayed if present.
        """
        return self._get_core_performance_attr('perf_name')

    @property
    def is_limited(self):
        """Boolean flag indicating limited availability."""
        return resolve_boolean(
            self._get_core_performance_attr('is_limited')
        )

    @property
    def cached_max_seats(self):
        """Maximum number of contiguous seats available."""
        return to_int_or_none(
            self._get_core_performance_attr('cached_max_seats')
        )

    @property
    def perf_type_code(self):
        """Type code, for internal use."""
        return self._get_core_performance_attr('perf_type_code')

    @property
    def ticket_types(self):
        """List of TicketType objects for this Performance."""
        if self._ticket_types is None:
            return self.get_availability()
        else:
            return self._ticket_types

    @ticket_types.setter
    def ticket_types(self, value):
        self._ticket_types = value

    @property
    def valid_ticket_quantities(self):
        """List of valid ticket quantities for this Performance."""
        if self._valid_ticket_quantities is None:

            crypto_block = self._get_date_time_options_crypto()

            resp_dict = self.get_core_api().availability_options(
                crypto_block=crypto_block,
                upfront_data_token=self.settings['upfront_data_token'],
                perf_token=self._perf_token,
                usage_date=date_to_yyyymmdd_or_none(self.usage_date),
                departure_date=date_to_yyyymmdd_or_none(self.departure_date),
                quantity_options_only=True
            )

            self._set_ticket_quantities_from_dict(
                resp_dict.get('quantity_options', None)
            )

        return self._valid_ticket_quantities

    def _set_ticket_quantities_from_dict(self, quantity_options_dict):

        valid_list = []

        if quantity_options_dict:

            valid_list = quantity_options_dict.get('valid_quantity', [])

            valid_list = [int(x) for x in valid_list]

        self.valid_ticket_quantities = valid_list

    @valid_ticket_quantities.setter
    def valid_ticket_quantities(self, value):
        self._valid_ticket_quantities = value

    @property
    def event(self):
        """The Event object this performance relates to."""
        if self._event is None:
            self.get_despatch_methods()

        if self._event:
            return self._event
        else:
            return None

    @event.setter
    def event(self, value):
        if value is None:
            self._event = False
        else:
            self._event = value

    def _get_date_time_options_crypto(self):

        crypto_block = self._get_crypto_block_for_object(
            method_name='date_time_options',
            interface_object=self._event,
        )

        if not crypto_block:
            self.event.get_performances()

            crypto_block = self._get_crypto_block_for_object(
                method_name='date_time_options',
                interface_object=self._event,
            )

        return crypto_block

    def get_availability(
        self, include_possible_concessions=None, no_of_tickets=None,
        include_available_seat_blocks=None, include_user_commission=None,
    ):
        """Retrieves ticket availability information for this Performance.

        Returns the list of TicketType objects, called internally by several
        methods, it calls the 'availability_options' API method.
        The 'ticket_types' property should be used to get this information,
        but this method can be called explicitly if required.

        Args:
            include_possible_concessions (boolean): Optional, flag to indicate
                whether to request possible_concession information.
            no_of_tickets (int): Optional, set the number of tickets to
                request, allows actual seats to be returned if possible.
            include_available_seat_blocks (boolean): Optional, flag to indicate
                whether to request the available seat blocks, allowing specific
                seats to be selected.
            include_user_commission (boolean): Optional, flag to indicate
                whether to request the user's commission information.

        Returns:
            list: List of TicketType objects
        """

        crypto_block = self._get_date_time_options_crypto()

        resp_dict = self.get_core_api().availability_options(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            perf_token=self._perf_token,
            usage_date=date_to_yyyymmdd_or_none(self.usage_date),
            departure_date=date_to_yyyymmdd_or_none(self.departure_date),
            self_print_mode='html', add_discounts=include_possible_concessions,
            no_of_tickets=no_of_tickets,
            add_free_seat_blocks=include_available_seat_blocks,
            add_user_commission=include_user_commission
        )

        self._set_crypto_block(
            crypto_block=resp_dict['crypto_block'],
            method_name='availability_options'
        )

        ticket_types = []

        if 'ticket_type' in resp_dict:
            for tt in resp_dict['ticket_type']:
                for pb in tt.price_bands:
                    ticket_type = availability.TicketType(
                        ticket_type_id=pb.band_token, core_ticket_type=tt,
                        core_price_band=pb,
                        core_currency=resp_dict.get('currency', None),
                        **self._internal_settings()
                    )

                    ticket_type.set_valid_quantities(
                        valid_quantities=resp_dict['quantity_options'][
                            'valid_quantity'
                        ]
                    )

                    ticket_types.append(ticket_type)

        self.ticket_types = ticket_types

        self._set_crypto_for_objects(
            crypto_block=resp_dict['crypto_block'],
            method_name='availability_options',
            interface_objects=ticket_types
        )

        self._set_ticket_quantities_from_dict(
            resp_dict.get('quantity_options', None)
        )

        despatch_methods = []

        if 'despatch_options' in resp_dict:

            for dm in resp_dict['despatch_options']['despatch_method']:
                despatch_method = availability.DespatchMethod(
                    despatch_id=dm.despatch_token,
                    core_despatch_method=dm,
                    core_currency=resp_dict.get('currency', None),
                    **self._internal_settings()
                )

                despatch_methods.append(despatch_method)

        self.despatch_methods = despatch_methods

        self._core_performance = resp_dict.get('performance')

        if 'event' in resp_dict:
            event = event_objs.Event(
                event_id=resp_dict['event'].event_id,
                core_event=resp_dict['event'],
                **self._internal_settings()
            )
            self.event = event

        return ticket_types

    def get_initial_ticket_type(self, no_of_tickets):
        if self.ticket_types:
            if int(no_of_tickets) <= 1:
                return self.ticket_types[0]
            else:
                for tt in self.ticket_types:
                    if int(tt.number_available) >= int(no_of_tickets):
                        return tt
        else:
            return None

    @property
    def ticket_types_by_type_then_price(self):
        """Ticket Type objects ordered by description then combined price."""
        return sorted(
            self.ticket_types, key=attrgetter(
                'description', 'price_combined_float'
            )
        )

    @property
    def ticket_types_by_cheapest_type(self):
        """Ticket Type objects ordered by cheapest description then price."""
        order_dict = {}

        i = 0
        for t in self.ticket_types_by_combined_price:
            if not order_dict or t.description not in order_dict:
                order_dict[t.description] = i
                i += 1

        return sorted(
            self.ticket_types, key=lambda x: (
                order_dict[x.description],
                x.price_combined_float,
                x.int_percentage_saving
            )
        )

    @property
    def ticket_types_by_combined_price(self):
        """Ticket Type objects ordered by combined price."""
        return sorted(
            self.ticket_types, key=attrgetter(
                'price_combined_float', 'int_percentage_saving'
            )
        )

    @property
    def ticket_types_by_saving(self):
        """Ticket Type objects ordered by percentage saving."""
        return sorted(
            self.ticket_types,
            key=attrgetter('int_percentage_saving', 'price_combined_float'),
            reverse=True
        )

    @property
    def unique_combined_prices(self):
        """List of unique combined prices."""
        unique_list = []

        for tt in self.ticket_types_by_combined_price:
            if tt.price_combined not in unique_list:
                unique_list.append(tt.price_combined)

        return unique_list

    @property
    def despatch_methods(self):
        """List of DespatchMethod objects for this Performance."""
        if self._despatch_methods is None:
            return self.get_despatch_methods()
        else:
            return self._despatch_methods

    @despatch_methods.setter
    def despatch_methods(self, value):
        self._despatch_methods = value

    def get_despatch_methods(self):
        """Retrieves despatch method information for this Performance.

        Returns the list of DespatchMethod objects, called internally by some
        methods, it calls the 'despatch_options' API method.
        The 'despatch_methods' property should be used to get this information,
        but this method can be called explicitly if required.

        Returns:
            list: List of DespatchMethod objects
        """

        crypto_block = self._get_date_time_options_crypto()

        resp_dict = self.get_core_api().despatch_options(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            perf_token=self._perf_token,
            usage_date=date_to_yyyymmdd_or_none(self.usage_date),
            departure_date=date_to_yyyymmdd_or_none(self.departure_date),
            self_print_mode='html'
        )

        despatch_methods = []

        for dm in resp_dict['despatch_options']['despatch_method']:
            despatch_method = availability.DespatchMethod(
                despatch_id=dm.despatch_token,
                core_despatch_method=dm,
                core_currency=resp_dict.get('currency', None),
                **self._internal_settings()
            )

            despatch_methods.append(despatch_method)

        self.despatch_methods = despatch_methods

        self._core_performance = resp_dict.get('performance')

        event = None
        if 'event' in resp_dict:
            event = event_objs.Event(
                event_id=resp_dict['event'].event_id,
                core_event=resp_dict['event'],
                **self._internal_settings()
            )
        self.event = event

        return despatch_methods



import logging

from pyticketswitch import settings as default_settings
from pyticketswitch.interface import CoreAPI
from pyticketswitch.util import (
    resolve_boolean, format_price_with_symbol,
    to_float_or_none, to_int_or_none,
)

logger = logging.getLogger(__name__)


class InterfaceObject(object):
    """Superclass for all objects that will perform API operations.

    The class should not be instantiated directly, it is designed to
    be subclassed. Contains a number of internal methods that provide
    standard functionality for API operations.

    Subclasses of this object should call the constructor of this
    object.

    Args:
        username (string): TSW user
        password (string): password for TSW user
        url (string): TSW API URL
        accept_language (string): user's HTTP Accept-Language header
        api_request_timeout (int): API timeout in seconds
        no_time_descr (string): text to use if no time is returned
            by the API for a performance
        default_concession_descr (string): text to use if no description
            is returned by the API for a concession
        remote_ip (string): user's IP address (internal use)
        remote_site (string): domain of the user request (internal use)
        ext_start_session_url (string): URL for start_session (internal user)
        additional_elements (dict): An optional dictionary of key/values to
            include in the XML passed to the API.
        upfront_data_token (string): upfront data token is required by the API
            in certain cases, such as for redeem
        requests_session (requests.Session object): optional Requests session
            to use for making HTTP requests
    """

    CRYPTO_PREFIX = 'CRYPTO_BLOCK'
    USERNAME_PREFIX = 'USERNAME'
    RUNNING_USER_PREFIX = 'RUNNING_USER'

    def __init__(self, **kwargs):

        self._session = None
        self._core_api = None
        self.settings = self._get_settings()
        self._session_store = {}

        if 'session' in kwargs:
            self._session = kwargs.pop('session')

        if '_core_api' in kwargs:
            self._core_api = kwargs.pop('_core_api')

        if '_settings' in kwargs:
            self.settings = kwargs.pop('_settings')

        if '_session_store' in kwargs:
            self._session_store = kwargs.pop('_session_store')

        if kwargs:
            self._configure(**kwargs)

    def _get_settings(
            self, username=None, password=None, sub_id=None, url=None,
            no_time_descr=None, api_request_timeout=None,
            default_concession_descr=None, remote_ip=None,
            remote_site=None, accept_language=None,
            ext_start_session_url=None,
            additional_elements=None, upfront_data_token=None,
            requests_session=None):

        return {
            'username': username,
            'password': password,
            'sub_id': sub_id,
            'remote_ip': remote_ip,
            'remote_site': remote_site,
            'accept_language': accept_language,
            'url': url,
            'ext_start_session_url': ext_start_session_url,
            'api_request_timeout': api_request_timeout,
            'no_time_descr': no_time_descr,
            'default_concession_descr': default_concession_descr,
            'additional_elements': additional_elements,
            'upfront_data_token': upfront_data_token,
            'requests_session': requests_session,
        }

    def _configure(
            self, username=None, password=None, sub_id=None, url=None,
            no_time_descr=None, api_request_timeout=None,
            default_concession_descr=None, remote_ip=None,
            remote_site=None, accept_language=None, ext_start_session_url=None,
            additional_elements=None, upfront_data_token=None,
            requests_session=None):

        if (not username) and remote_ip and remote_site:
            username = self._get_cached_username(
                remote_ip=remote_ip,
                remote_site=remote_site
            )

        if (
            (
                'username' in self.settings and
                self.settings['username'] and
                username != self.settings['username']
            ) or
            (
                'remote_ip' in self.settings and
                self.settings['remote_ip'] and
                remote_ip != self.settings['remote_ip']
            ) or
            (
                'remote_site' in self.settings and
                self.settings['remote_site'] and
                remote_site != self.settings['remote_site']
            )
        ):
            self._clear_crypto_blocks()

        if not url:
            url = default_settings.API_URL

        if not ext_start_session_url:
            ext_start_session_url = default_settings.EXT_START_SESSION_URL

        if not api_request_timeout:
            api_request_timeout = default_settings.API_REQUEST_TIMEOUT

        if not no_time_descr:
            no_time_descr = default_settings.NO_TIME_DESCR

        if not default_concession_descr:
            default_concession_descr = (
                default_settings.DEFAULT_CONCESSION_DESCR
            )

        self.settings = self._get_settings(
            username=username, password=password, sub_id=sub_id,
            url=url, no_time_descr=no_time_descr,
            api_request_timeout=api_request_timeout,
            default_concession_descr=default_concession_descr,
            remote_ip=remote_ip,
            remote_site=remote_site,
            accept_language=accept_language,
            ext_start_session_url=ext_start_session_url,
            additional_elements=additional_elements,
            upfront_data_token=upfront_data_token,
            requests_session=requests_session,
        )

        self._core_api = CoreAPI(
            username=username,
            password=password,
            sub_id=sub_id,
            url=url,
            remote_ip=remote_ip,
            remote_site=remote_site,
            accept_language=accept_language,
            ext_start_session_url=ext_start_session_url,
            api_request_timeout=api_request_timeout,
            additional_elements=additional_elements,
            requests_session=requests_session,
        )

    def get_core_api(self):
        return self._core_api

    def set_session(self, session):
        self._session = session

    def _password_is_set(self):

        if self.settings.get('password'):
            return True
        else:
            return False

    def _start_session(self):
        crypto_block = self.get_core_api().start_session()

        username = self.get_core_api().username

        if (
            not self.settings['username'] or
            username != self.settings['username']
        ):

            remote_ip = self.settings['remote_ip']
            remote_site = self.settings['remote_site']

            self.settings['username'] = username
            self._set_cached_username(
                username=username,
                remote_ip=remote_ip,
                remote_site=remote_site
            )

            self._set_cached_running_user(
                username=username,
                running_user=self.get_core_api().running_user,
            )

        if crypto_block:
            self._set_crypto_block(
                crypto_block=crypto_block,
                method_name='start_session'
            )

        return crypto_block

    def get_username(self):

        if not self.settings['username']:

            remote_ip = self.settings['remote_ip']
            remote_site = self.settings['remote_site']

            username = self._get_cached_username(
                remote_ip=remote_ip,
                remote_site=remote_site
            )

            if not username:
                self._start_session()

        return self.settings['username']

    def _get_running_user(self):
        if not self.get_core_api().running_user:

            running_user = self._get_cached_running_user(
                username=self.get_username()
            )

            if running_user:
                self.get_core_api().running_user = running_user
            else:
                self._start_session()

        return self.get_core_api().running_user

    def get_restrict_group(self):
        return self._get_running_user().sphinx_restrict_group

    def get_default_language_code(self):
        running_user = self._get_running_user()
        if running_user:
            return running_user.default_lang_code

    def get_content_language(self):
        return self.get_core_api().content_language

    def _store_data(self, key, data, save_session=True):

        logger.debug('_store_data, key: %s, data: %s', key, data)

        self._session_store[key] = data

        if self._session is not None:
            self._session[key] = data

            if save_session and hasattr(self._session, 'save'):
                self._session.save()

    def _retrieve_data(self, key):

        data = self._session_store.get(key, None)

        if not data and self._session is not None:
            data = self._session.get(key)

        logger.debug('_retrieve_data, key: %s, data: %s', key, data)

        return data

    def _get_username_session_key(self, remote_ip, remote_site):
        return '{0}_{1}_{2}'.format(
            self.USERNAME_PREFIX, remote_ip, remote_site
        )

    def _get_cached_username(
            self, remote_ip, remote_site):
        key = self._get_username_session_key(
            remote_ip=remote_ip,
            remote_site=remote_site
        )

        return self._retrieve_data(key)

    def _set_cached_username(
            self, username, remote_ip, remote_site):
        key = self._get_username_session_key(
            remote_ip=remote_ip,
            remote_site=remote_site
        )

        self._store_data(key=key, data=username)

    def _get_running_user_session_key(self, username):
        return '{0}_{1}'.format(
            self.RUNNING_USER_PREFIX, username
        )

    def _get_cached_running_user(self, username):

        key = self._get_running_user_session_key(
            username=username,
        )

        return self._retrieve_data(key)

    def _set_cached_running_user(self, username, running_user):

        key = self._get_running_user_session_key(
            username=username,
        )

        self._store_data(key=key, data=running_user)

    def _get_crypto_session_key(self, username, method_name):
        return '{0}_{1}_{2}'.format(
            self.CRYPTO_PREFIX, username, method_name
        )

    def _clear_crypto_blocks(self):

        logger.debug('_clear_crypto_blocks called')

        for key in self._session_store.keys():
            if key.startswith(self.CRYPTO_PREFIX):
                del self._session_store[key]

        if self._session is not None:

            if hasattr(self._session, 'flush_crypto_blocks'):
                self._session.flush_crypto_blocks()
            else:
                for key in self._session.keys():
                    if key.startswith(self.CRYPTO_PREFIX):
                        del self._session[key]

    def get_crypto_block(
            self, method_name, password_required=True):

        crypto_block = None

        if (
            password_required or (
                not password_required and not self._password_is_set()
            )
        ):
            if not self.settings.get('username', False):
                start_session_crypto = self._start_session()
            else:
                start_session_crypto = None

            session_key = self._get_crypto_session_key(
                username=self.settings['username'],
                method_name=method_name
            )

            crypto_block = self._retrieve_data(session_key)

            if not crypto_block and method_name == 'start_session':

                if start_session_crypto:
                    crypto_block = start_session_crypto
                else:
                    crypto_block = self._start_session()

        return crypto_block

    def _set_crypto_block(self, crypto_block, method_name):

        session_key = self._get_crypto_session_key(
            username=self.settings['username'], method_name=method_name
        )

        self._store_data(key=session_key, data=crypto_block)

    def _get_crypto_object_key(
            self, username, method_name, interface_object):

        return '{0}_{1}'.format(
            self._get_crypto_session_key(
                username=username, method_name=method_name
            ), interface_object._get_cache_key()
        )

    def _set_crypto_for_objects(
            self, crypto_block, method_name, interface_objects):

        for i, obj in enumerate(interface_objects):
            key = self._get_crypto_object_key(
                username=self.settings['username'],
                method_name=method_name,
                interface_object=obj
            )

            if i == len(interface_objects) - 1:
                save_session = True
            else:
                save_session = False

            self._store_data(
                key=key, data=crypto_block,
                save_session=save_session
            )

    def _set_crypto_for_object(
            self, crypto_block, method_name, interface_object):

        key = self._get_crypto_object_key(
            username=self.settings['username'],
            method_name=method_name,
            interface_object=interface_object
        )

        self._store_data(
            key=key, data=crypto_block,
        )

    def _get_crypto_block_for_object(self, method_name, interface_object):

        if not self.settings.get('username', False):
            self._start_session()

        key = self._get_crypto_object_key(
            username=self.settings['username'],
            method_name=method_name,
            interface_object=interface_object
        )

        return self._retrieve_data(key)

    def _internal_settings(self):
        return {
            'session': self._session,
            '_core_api': self._core_api,
            '_settings': self.settings,
            '_session_store': self._session_store,
        }

    def __getstate__(self):

        d = self.__dict__.copy()
        d['_session'] = None
        d['_core_api'] = None
        d['settings'] = self._get_settings()
        d['_session_store'] = {}

        return d


class Currency(object):
    """Represents a Currency in TSW, used in several other objects.

    The constructor is for internal user only.
    """

    def __init__(
        self,
        core_currency
    ):
        self._core_currency = core_currency

    @property
    def code(self):
        return self._core_currency.currency_code

    @property
    def number(self):
        return to_int_or_none(
            self._core_currency.currency_number
        )

    @property
    def pre_symbol(self):
        return self._core_currency.currency_pre_symbol

    @property
    def post_symbol(self):
        return self._core_currency.currency_post_symbol

    @property
    def factor(self):
        return to_int_or_none(
            self._core_currency.currency_factor
        )

    @property
    def decimal_places(self):
        return to_int_or_none(
            self._core_currency.currency_places
        )


class Seat(object):
    """Represents a Seat in TSW, used in several other objects.

    The constructor is for internal user only.
    """

    def __init__(
        self,
        core_seat
    ):
        self._core_seat = core_seat

    @property
    def seat_id(self):
        return self._core_seat.full_id

    @property
    def is_restricted_view(self):
        """Boolean representing whether the seat has a restricted view."""
        return resolve_boolean(self._core_seat.is_restricted_view)

    @property
    def seat_text(self):
        """Additional information about this seat.

        E.g. restricted legroom
        """
        return self._core_seat.seat_text

    @property
    def column_id(self):
        return self._core_seat.col_id

    @property
    def row_id(self):
        return self._core_seat.row_id

    @property
    def separator(self):
        """The seat id is made up of row_id + separator + column_id.
        Returns empty string if there is no separator.
        """
        return self._core_seat.separator or ''

    @property
    def column_sort_id(self):
        col = self._core_seat.col_id

        if col:
            try:
                col = int(col)
            except ValueError:
                pass

        return col

    @property
    def row_sort_id(self):
        row = self._core_seat.row_id

        if row:
            try:
                row = int(row)
            except ValueError:
                pass

        return row

    @property
    def barcode(self):
        return self._core_seat.barcode


class SeatBlock(object):
    """Represents a block of seats in TSW, used when selecting seats in the
    booking process.

    Args:
        seat_block_id (string): Id of the SeatBlock
    """

    def __init__(
        self,
        seat_block_id=None,
        core_seat_block=None,
    ):
        self._core_seat_block = core_seat_block

        if not seat_block_id and core_seat_block:
            seat_block_id = core_seat_block.seat_block_token

        self.seat_block_id = seat_block_id
        self._seats = False

    @property
    def block_length(self):
        if self._core_seat_block:
            return to_int_or_none(
                self._core_seat_block.block_length
            )
        return None

    @property
    def seats(self):
        if self._seats is False:
            if self._core_seat_block:

                self._seats = []

                for seat in self._core_seat_block.seats:
                    self._seats.append(
                        Seat(core_seat=seat)
                    )
            else:
                self._seats = None

        return self._seats


class Customer(object):
    """Object that represents a customer.

    The 'core_customer' argument in the constructor is for internal use.

    The user, supplier and world can_use_data arguments are for data
    protection purposes and control who can use the customer's data.
    User is the TSW affiliate, supplier is the backend ticket supplier and
    world is third parties.

    Args:
        first_name (string): First name.
        last_name (string): Last name.
        home_phone (string): Home phone number.
        work_phone (string): Work phone number.
        address (Address): Home address.
        title (string): Optional, title.
        email_address (string): Optional, email address.
        user_can_use_data (boolean): Optional, data protection flag.
        supplier_can_use_data (boolean): Optional, data protection flag.
        world_can_use_data (boolean): Optional, data protection flag.
    """

    def __init__(
        self,
        core_customer=None,
        first_name=None,
        last_name=None,
        home_phone=None,
        work_phone=None,
        address=None,
        title=None,
        email_address=None,
        user_can_use_data=None,
        supplier_can_use_data=None,
        world_can_use_data=None,
        languages=None,
    ):

        self._core_customer = core_customer
        self._first_name = first_name
        self._last_name = last_name
        self._home_phone = home_phone
        self._work_phone = work_phone
        self._address = address
        self._title = title
        self._email_address = email_address
        self._user_can_use_data = user_can_use_data
        self._supplier_can_use_data = supplier_can_use_data
        self._world_can_use_data = world_can_use_data

        self.languages = languages

    def _get_dict(self):
        return {
            'first_name': self.first_name,
            'last_name': self.last_name,
            'home_phone': self.home_phone,
            'work_phone': self.work_phone,
            'title': self.title,
            'email_address': self.email_address,
            'country_code': self.country_code,
            'address_line_one': self.address_line_one,
            'address_line_two': self.address_line_two,
            'town': self.town,
            'county': self.county,
            'postcode': self.postcode,
            'user_can_use_data': self.user_can_use_data,
            'supplier_can_use_data': self.supplier_can_use_data,
            'world_can_use_data': self.world_can_use_data,
        }

    @property
    def first_name(self):
        if self._core_customer:
            return self._core_customer.first_name
        else:
            return self._first_name

    @property
    def last_name(self):
        if self._core_customer:
            return self._core_customer.last_name
        else:
            return self._last_name

    @property
    def home_phone(self):
        if self._core_customer:
            return self._core_customer.home_phone
        else:
            return self._home_phone

    @property
    def work_phone(self):
        if self._core_customer:
            return self._core_customer.work_phone
        else:
            return self._work_phone

    @property
    def country_code(self):
        if self._core_customer:
            return self._core_customer.country_code
        else:
            return self._address.country_code

    @property
    def address_line_one(self):
        if self._core_customer:
            return self._core_customer.addr_line_one
        else:
            return self._address.address_line_one

    @property
    def address_line_two(self):
        if self._core_customer:
            return self._core_customer.addr_line_two
        else:
            return self._address.address_line_two

    @property
    def town(self):
        if self._core_customer:
            return self._core_customer.town
        else:
            return self._address.town

    @property
    def county(self):
        if self._core_customer:
            return self._core_customer.county
        else:
            return self._address.county

    @property
    def postcode(self):
        if self._core_customer:
            return self._core_customer.postcode
        else:
            return self._address.postcode

    @property
    def country(self):
        if self._core_customer:
            return self._core_customer.country
        else:
            return self._country

    @property
    def title(self):
        if self._core_customer:
            return self._core_customer.title
        else:
            return self._title

    @property
    def email_address(self):
        if self._core_customer:
            return self._core_customer.email_addr
        else:
            return self._email_address

    @property
    def user_can_use_data(self):
        if self._core_customer:
            return resolve_boolean(
                self._core_customer.dp_user
            )
        else:
            if isinstance(self._user_can_use_data, bool):
                return self._user_can_use_data
            elif isinstance(self._user_can_use_data, basestring):
                return resolve_boolean(
                    self._user_can_use_data
                )

    @property
    def supplier_can_use_data(self):
        if self._core_customer:
            return resolve_boolean(
                self._core_customer.dp_supplier
            )
        else:
            if isinstance(self._supplier_can_use_data, bool):
                return self._supplier_can_use_data
            elif isinstance(self._supplier_can_use_data, basestring):
                return resolve_boolean(
                    self._supplier_can_use_data
                )

    @property
    def world_can_use_data(self):
        if self._core_customer:
            return resolve_boolean(
                self._core_customer.dp_world
            )
        else:
            if isinstance(self._world_can_use_data, bool):
                return self._world_can_use_data
            elif isinstance(self._world_can_use_data, basestring):
                return resolve_boolean(
                    self._world_can_use_data
                )


class Card(object):
    """Object that represents a payment card.

    Args:
        card_number (string): The main card number.
        expiry_date (datetime.date): Expiry date of the card.
        cv_two (string): The card verification number (a.k.a CVV).
        start_date (datetime.date): Optional, start date of card.
        issue_number (string): Optional, card issue number.
        billing_address (Address): Optional, alternative billing address.
    """

    def __init__(
        self,
        card_number,
        expiry_date,
        cv_two,
        start_date=None,
        issue_number=None,
        billing_address=None
    ):

        self._card_number = card_number
        self._start_date = start_date
        self._expiry_date = expiry_date
        self._cv_two = cv_two
        self._issue_number = issue_number
        self._billing_address = billing_address

    def _get_dict(self):
        return {
            'card_number': self.card_number,
            'start_date': self.start_date_mmyy,
            'expiry_date': self.expiry_date_mmyy,
            'cv_two': self.cv_two,
            'issue_number': self.issue_number,
            'address_line_one': self.billing_address_line_one,
            'address_line_two': self.billing_address_line_two,
            'town': self.billing_town,
            'county': self.billing_county,
            'postcode': self.billing_postcode,
            'country_code': self.billing_country_code
        }

    @property
    def card_number(self):
        return self._card_number

    @property
    def start_date(self):
        return self._start_date

    @property
    def start_date_mmyy(self):
        """Returns string value in mmyy format"""
        if self.start_date:
            return self.start_date.strftime('%m%y')
        else:
            return None

    @property
    def expiry_date(self):
        return self._expiry_date

    @property
    def expiry_date_mmyy(self):
        """Returns string value in mmyy format"""
        return self.expiry_date.strftime('%m%y')

    @property
    def cv_two(self):
        return self._cv_two

    @property
    def issue_number(self):
        return self._issue_number

    @property
    def billing_address_line_one(self):
        if self._billing_address:
            return self._billing_address.address_line_one
        else:
            return None

    @property
    def billing_address_line_two(self):
        if self._billing_address:
            return self._billing_address.address_line_two
        else:
            return None

    @property
    def billing_town(self):
        if self._billing_address:
            return self._billing_address.town
        else:
            return None

    @property
    def billing_county(self):
        if self._billing_address:
            return self._billing_address.county
        else:
            return None

    @property
    def billing_postcode(self):
        if self._billing_address:
            return self._billing_address.postcode
        else:
            return None

    @property
    def billing_country_code(self):
        if self._billing_address:
            return self._billing_address.country_code
        else:
            return None


class Address(object):
    """Object that represents an address.

    The attributes are the same as the constructor arguments.

    Args:
        address_line_one (string): 1st line of address.
        country_code (string): 2 digit ISO 3166 country code.
        address_line_two (string): Optional, 2nd line of address.
        town (string): Optional, town.
        county (string): Optional, county.
        postcode (string): Optional, postcode.
    """

    def __init__(
        self,
        address_line_one,
        country_code,
        address_line_two=None,
        town=None,
        county=None,
        postcode=None
    ):

        self.address_line_one = address_line_one
        self.address_line_two = address_line_two
        self.town = town
        self.county = county
        self.postcode = postcode
        self.country_code = country_code


class SpecialOffer(object):
    """Represents a special offer in TSW, used in objects with cost range
    information.

    The constructor is for internal user only.
    """

    def __init__(self, core_offer, core_currency, offer_type):

        self._core_offer = core_offer
        self._core_currency = core_currency
        self.offer_type = offer_type
        self.currency = Currency(
            core_currency=core_currency
        )

    @property
    def full_combined_price(self):
        """Formatted string value of the full combined price with
        currency symbol.
        """
        return format_price_with_symbol(
            self._core_offer['full_combined'],
            self.currency.pre_symbol,
            self.currency.post_symbol
        )

    @property
    def full_combined_price_float(self):
        """Float value of the full combined price."""
        return to_float_or_none(
            self._core_offer['full_combined']
        )

    @property
    def full_seatprice(self):
        """Formatted string value of the full seatprice price with
        currency symbol.
        """
        return format_price_with_symbol(
            self._core_offer['full_seatprice'],
            self.currency.pre_symbol,
            self.currency.post_symbol
        )

    @property
    def full_seatprice_float(self):
        """Float value of the full seatprice price."""
        return to_float_or_none(
            self._core_offer['full_seatprice']
        )

    @property
    def full_surcharge_price(self):
        """Formatted string value of the full surcharge price with
        currency symbol.
        """
        return format_price_with_symbol(
            self._core_offer['full_surcharge'],
            self.currency.pre_symbol,
            self.currency.post_symbol
        )

    @property
    def full_surcharge_price_float(self):
        """Float value of the full surcharge price."""
        return to_float_or_none(
            self._core_offer['full_surcharge']
        )

    @property
    def offer_combined_price(self):
        """Formatted string value of the offer combined price with
        currency symbol.
        """
        return format_price_with_symbol(
            self._core_offer['offer_combined'],
            self.currency.pre_symbol,
            self.currency.post_symbol
        )

    @property
    def offer_combined_price_float(self):
        """Float value of the offer combined price."""
        return to_float_or_none(
            self._core_offer['offer_combined']
        )

    @property
    def offer_seatprice(self):
        """Formatted string value of the offer seatprice price with
        currency symbol.
        """
        return format_price_with_symbol(
            self._core_offer['offer_seatprice'],
            self.currency.pre_symbol,
            self.currency.post_symbol
        )

    @property
    def offer_seatprice_float(self):
        """Float value of the offer seatprice price."""
        return to_float_or_none(
            self._core_offer['offer_seatprice']
        )

    @property
    def offer_surcharge_price(self):
        """Formatted string value of the offer surcharge price with
        currency symbol.
        """
        return format_price_with_symbol(
            self._core_offer['offer_surcharge'],
            self.currency.pre_symbol,
            self.currency.post_symbol
        )

    @property
    def offer_surcharge_price_float(self):
        """Float value of the offer surcharge price."""
        return to_float_or_none(
            self._core_offer['offer_surcharge']
        )

    @property
    def percentage_saving(self):
        """Formatted string value of the percentage saving with
        percent sign.
        """
        return '{0}%'.format(
            self._core_offer['percentage_saving']
        )

    @property
    def percentage_saving_float(self):
        """Float value of the percentage saving."""
        return to_float_or_none(
            self._core_offer['percentage_saving']
        )

    @property
    def absolute_saving_float(self):
        """Float value of the absolute saving."""
        ab_val = self._core_offer.get('absolute_saving', None)

        if (
            ab_val is None and
            self.full_combined_price_float is not None and
            self.offer_combined_price_float is not None
        ):
            ab_val = (
                self.full_combined_price_float -
                self.offer_combined_price_float
            )

        return to_float_or_none(ab_val)

    @property
    def absolute_saving(self):
        """Formatted string value of the absolute saving value with
        currency symbol.
        """

        ret_val = None

        if self.absolute_saving_float is not None:

            ret_val = format_price_with_symbol(
                self.absolute_saving_float,
                self.currency.pre_symbol,
                self.currency.post_symbol
            )
        return ret_val

    @property
    def is_no_booking_fee_offer(self):
        """Boolean indicating if this is a no booking fee offer."""
        if (
            self.full_seatprice_float == self.offer_seatprice_float and
            (
                self.full_combined_price_float >
                self.offer_combined_price_float
            ) and
            not self.offer_surcharge_price_float
        ):
            return True

        return False


class CostRangeMixin(object):
    """Object to provide common cost range related functionality."""

    def _get_core_cost_range(self):
        """ This method must be overridden and should return a core
            cost range object.
        """
        raise NotImplementedError(
            'Subclasses must override _get_core_cost_range()'
        )

    @property
    def special_offers(self):
        """Returns a list of SpecialOffer objects."""
        offers = [
            self.best_value_offer, self.max_saving_offer,
            self.top_price_offer
        ]

        return [o for o in offers if o is not None]

    @property
    def currency(self):
        """Returns a currency object."""
        currency = None
        cost_range = self._get_core_cost_range()

        if cost_range:
            currency = Currency(
                core_currency=cost_range.currency
            )

        return currency

    @property
    def best_value_offer(self):
        """Object representing the best value offer. Returns None
        if there is no best value offer.
        """
        offer = None
        cost_range = self._get_core_cost_range()

        if cost_range and cost_range.best_value_offer:

            offer = SpecialOffer(
                core_offer=cost_range.best_value_offer,
                core_currency=cost_range.currency,
                offer_type='best_value',
            )

        return offer

    @property
    def max_saving_offer(self):
        """Object representing the max saving offer. Returns None
        if there is no max saving offer.
        """
        offer = None
        cost_range = self._get_core_cost_range()

        if cost_range and cost_range.max_saving_offer:

            offer = SpecialOffer(
                core_offer=cost_range.max_saving_offer,
                core_currency=cost_range.currency,
                offer_type='max_saving',
            )

        return offer

    @property
    def top_price_offer(self):
        """Object representing the top price offer. Returns None
        if there is no top price offer.
        """
        offer = None
        cost_range = self._get_core_cost_range()

        if cost_range and cost_range.top_price_offer:

            offer = SpecialOffer(
                core_offer=cost_range.top_price_offer,
                core_currency=cost_range.currency,
                offer_type='top_price',
            )

        return offer

    @property
    def min_seatprice(self):
        """Formatted string value of the minimun seat price with
        currency symbol.
        """
        min_price = None
        cost_range = self._get_core_cost_range()

        if cost_range:

            min_price = format_price_with_symbol(
                cost_range.min_seatprice,
                cost_range.currency.currency_pre_symbol,
                cost_range.currency.currency_post_symbol
            )

        return min_price

    @property
    def min_seatprice_float(self):
        """Float value of the minumum seat price."""
        fl_sp = None

        cost_range = self._get_core_cost_range()

        if cost_range:
            fl_sp = to_float_or_none(
                cost_range.min_seatprice
            )
        return fl_sp

    @property
    def min_combined_price(self):
        """Formatted string value of the minimun combined price with
        currency symbol.
        """
        min_price = None
        cost_range = self._get_core_cost_range()

        if cost_range:

            min_price = format_price_with_symbol(
                cost_range.min_combined,
                cost_range.currency.currency_pre_symbol,
                cost_range.currency.currency_post_symbol
            )

        return min_price

    @property
    def min_combined_price_float(self):
        """Float value of the minumum combined price."""
        fl_sp = None

        cost_range = self._get_core_cost_range()

        if cost_range:
            fl_sp = to_float_or_none(
                cost_range.min_combined
            )
        return fl_sp

    @property
    def max_combined_price_float(self):
        """Float value of the maximum combined price."""
        fl_sp = None

        cost_range = self._get_core_cost_range()

        if cost_range:
            fl_sp = to_float_or_none(
                cost_range.max_combined
            )
        return fl_sp

    @property
    def is_special_offer(self):
        """Boolean indicating if the object has a special offer."""
        if self.special_offers:
            return True

        return False

    @property
    def max_saving_percent(self):
        """Formatted string value of the maximum saving percentage
        with a '%' symbol.
        """
        if self.best_value_offer:
            return self.best_value_offer.percentage_saving

        return None

    @property
    def max_saving_absolute(self):
        """Formatted string value of the maximum possible saving with
        currency symbol.
        """
        if self.max_saving_offer:
            return self.max_saving_offer.absolute_saving

        return None

    @property
    def best_value_non_offer_combined_price(self):
        """Formatted string value of the original cost of the best value
        offer price with currency symbol (i.e. if there was no offer).
        """
        if self.best_value_offer:
            return self.best_value_offer.full_combined_price

        return None

    @property
    def best_value_offer_combined_price(self):
        """Formatted string value of the best value offer price with
        currency symbol.
        """
        if self.best_value_offer:
            return self.best_value_offer.offer_combined_price

        return None

    @property
    def has_no_booking_fee_offer(self):
        """Boolean to indicate if there is an offer that has no
        booking fee."""
        for o in self.special_offers:
            if o.is_no_booking_fee_offer:
                return True

        return False

    @property
    def cached_valid_ticket_quantities(self):
        """Returns a list of the valid ticket quantities that have been
        cached"""
        ticket_quantities = []
        cost_range = self._get_core_cost_range()

        if cost_range:
            if cost_range.quantity_options:
                ticket_quantities = cost_range.quantity_options.get(
                    'valid_quantity', []
                )
                ticket_quantities = [int(x) for x in ticket_quantities]

        return ticket_quantities

    @property
    def no_singles(self):
        """Returns a sub cost range object representing a cost range when
        selecting more than one ticket."""

        cost_range = self._get_core_cost_range()

        if cost_range and cost_range.no_singles_cost_range:
            return CostRange(cost_range.no_singles_cost_range)

        return None


class CostRange(CostRangeMixin):
    """A class providing the same functionality as CostRangeMixin, for use
    where an actual cost range object is required"""

    def __init__(self, core_cost_range):

        self._core_cost_range = core_cost_range

    def _get_core_cost_range(self):
        return self._core_cost_range


class Commission(object):
    """Object that represents a commission (either user comission or gross
    commission)
    """

    def __init__(self, core_commission):
        self._core_commission = core_commission

    @property
    def amount_excluding_vat(self):
        """Float value of the commission excluding VAT"""
        if self._core_commission.amount_excluding_vat:
            return to_float_or_none(
                self._core_commission.amount_excluding_vat
            )
        return None

    @property
    def amount_including_vat(self):
        """Float value of the commission including VAT"""
        if self._core_commission.amount_including_vat:
            return to_float_or_none(
                self._core_commission.amount_including_vat
            )
        return None

    @property
    def commission_currency(self):
        """Returns Currency object representing the currency that the
        commission values are provided in
        """
        if self._core_commission.commission_currency:
            return Currency(
                core_currency=self._core_commission.commission_currency
            )
        return None



from setuptools import setup

setup(
    name='pyticketswitch',
    version='1.8.3',
    author='Ingresso',
    author_email='systems@ingresso.co.uk',
    url='https://github.com/ingresso-group/pyticketswitch/',
    packages=[
        'pyticketswitch',
        'pyticketswitch.interface_objects'
    ],
    license='MIT',
    description='A Python interface for the Ingresso XML Core API',
    long_description=open('README.rst').read(),
    install_requires=[
        'requests>=2.0.0',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.7',
        'Intended Audience :: Developers',
        'Natural Language :: English',
    ],
)



import requests
try:
    import xml.etree.cElementTree as xml
except ImportError:
    import xml.etree.ElementTree as xml
from datetime import datetime
import logging

from util import create_xml_from_dict, dict_ignore_nones
from api_exceptions import CommsException, InvalidResponse
import parse
import settings

logger = logging.getLogger(__name__)
filelog = logging.getLogger('filelog.' + __name__)


class CoreAPI(object):

    def __init__(
            self, username, password, url,
            remote_ip, remote_site, accept_language,
            ext_start_session_url, api_request_timeout,
            sub_id=None,
            additional_elements=None,
            requests_session=None):

        self.username = username
        self.password = password
        self.sub_id = sub_id

        self.url = url
        self.remote_ip = remote_ip
        self.remote_site = remote_site
        self.accept_language = accept_language
        self.ext_start_session_url = ext_start_session_url
        self.content_language = None
        self.running_user = None

        if api_request_timeout:
            self.api_request_timeout = api_request_timeout
        else:
            self.api_request_timeout = settings.API_REQUEST_TIMEOUT

        if not additional_elements:
            additional_elements = {}
        self.additional_elements = additional_elements

        # If no Requests session create a new one
        if not requests_session:
            requests_session = requests.Session()
        self.requests_session = requests_session

    def _post(self, method_name, data, url, headers=None):

        filelog.debug(
            u'URL=%s; API_REQUEST=%s',
            url, unicode(data, 'UTF-8')
        )

        response_string = None

        before = datetime.now()
        after = None

        try:
            response = self.requests_session.post(
                url=url, data=data, headers=headers,
                timeout=self.api_request_timeout,
            )
            response.raise_for_status()

        except requests.exceptions.HTTPError as e:
            after = datetime.now()
            raise CommsException(
                underlying_exception=e,
                description=(
                    'HTTPError, message={0}'.format(
                        getattr(e, 'message', None),
                    )
                )
            )

        except requests.exceptions.Timeout as e:
            after = datetime.now()
            raise CommsException(
                underlying_exception=e,
                description=(
                    'Timeout, message={0}'.format(
                        getattr(e, 'message', None),
                    )
                )
            )

        except requests.exceptions.ConnectionError as e:
            after = datetime.now()
            raise CommsException(
                underlying_exception=e,
                description=(
                    'ConnectionError, message={0}'.format(
                        getattr(e, 'message', None),
                    )
                )
            )

        except requests.exceptions.RequestException as e:
            after = datetime.now()
            raise CommsException(
                underlying_exception=e,
                description=(
                    'RequestException, message={0}'.format(
                        getattr(e, 'message', None),
                    )
                )
            )

        else:
            after = datetime.now()
            response_string = response.content
            self.content_language = response.headers.get(
                'Content-Language')

            filelog.debug(
                u'API_RESPONSE=%s',
                unicode(response_string, 'UTF-8')
            )

        finally:
            if after:
                time_taken = (after - before).total_seconds()
                logger.debug(
                    'url=%s, api_call=%s, time_taken=%s',
                    url, method_name, time_taken
                )
            else:
                logger.error(
                    (
                        'Response time unknown, url=%s, ' +
                        'api_call=%s, request_time=%s'
                    ),
                    url, method_name, before.strftime('%d/%m/%Y %H:%M:%S')
                )

        return response_string

    def _create_xml_and_post(self, method_name, arg_dict, url=None):

        data = xml.tostring(
            create_xml_from_dict(method_name, arg_dict),
            encoding='UTF-8'
        )

        if not url:
            url = self.url

        headers = {
            'Content-Type': 'text/xml',
        }

        if self.accept_language:
            headers['Accept-Language'] = self.accept_language

        try:
            response = xml.fromstring(
                self._post(
                    method_name=method_name,
                    data=data,
                    headers=headers,
                    url=url
                )
            )
        except CommsException as e:
            logger.error(e)
            raise e

        except xml.ParseError as e:

            err_string = 'XML parsing error, detail="{0}", arguments="{1}"'

            raise InvalidResponse(
                underlying_exception=e,
                description=err_string.format(
                    str(e), arg_dict
                ),
            )

        return response

    def make_core_request(self, api_call, **kwargs):

        args = {
            'user_id': self.username,
            'sub_id': self.sub_id,
            'remote_ip': self.remote_ip,
            'remote_site': self.remote_site,
        }

        args.update(self.additional_elements)

        args.update(kwargs)

        return self._create_xml_and_post(
            method_name=api_call,
            arg_dict=dict_ignore_nones(**args),
            url=self.url
        )

    def parse_response(self, parse_function, xml_elem):
        """ Calls the specified parse function

        Calls the specified parse function and logs
        any errors that are raised.

        """
        try:
            result = parse_function(
                parse.script_error(xml_elem)
            )
        except Exception as e:
            logger.error(e)
            raise e

        return result

    def start_session_resolve_user(
            self, user_id=None, remote_site=None, remote_ip=None):

        resp = self._create_xml_and_post(
            method_name='start_session',
            arg_dict=dict_ignore_nones(
                user_id=user_id,
                remote_site=remote_site,
                remote_ip=remote_ip
            ),
            url=self.ext_start_session_url
        )

        return self.parse_response(
            parse.start_session_resolve_user_result, resp
        )

    def start_session(self):

        if not self.password or not self.username:
            resp = self.start_session_resolve_user(
                user_id=self.username,
                remote_ip=self.remote_ip,
                remote_site=self.remote_site,
            )

            crypto_block = resp['crypto_block']
            self.username = resp['running_user'].user_id
            self.running_user = resp['running_user']

        else:
            arg_dict={
                'user_id': self.username,
                'user_passwd': self.password,
            }

            if self.sub_id:
                arg_dict.update(
                    sub_id=self.sub_id,
                )

            resp = self._create_xml_and_post(
                method_name='start_session',
                arg_dict=arg_dict,
            )

            parse.script_error(resp)
            running_user = resp.find('running_user')

            if running_user is not None:
                self.running_user = parse._parse_running_user(running_user)

            crypto_block = resp.findtext('crypto_block')

        return crypto_block

    def style_map(self, map_key):
        resp = self.make_core_request(
            'style_map',
            user_passwd=self.password,
            map_key=map_key
        )

        return self.parse_response(parse.style_map_result, resp)

    def event_search(
            self, crypto_block=None, upfront_data_token=None, s_keys=None,
            s_dates=None, s_coco=None, s_city=None, s_geo=None, s_geo_lat=None,
            s_geo_long=None, s_geo_rad_km=None, s_src=None, s_area=None,
            s_ven=None, s_eve=None, s_class=None, event_token_list=None,
            request_source_info=None, request_extra_info=None,
            request_video_iframe=None, request_cost_range=None,
            request_media=None, request_custom_fields=None, request_reviews=None,
            request_avail_details=None,
            s_top=None, s_user_rating=None,
            s_critic_rating=None, s_auto_range=None, page_length=None,
            page_number=None,
            s_cust_fltr=None, s_airport=None,
            mime_text_type=None):
        if crypto_block is None:
            user_passwd = self.password
        else:
            user_passwd = None

        resp = self.make_core_request(
            'event_search',
            user_passwd=user_passwd, crypto_block=crypto_block,
            upfront_data_token=upfront_data_token, s_keys=s_keys,
            s_dates=s_dates, s_coco=s_coco, s_geo=s_geo, s_geo_lat=s_geo_lat,
            s_geo_long=s_geo_long, s_geo_rad_km=s_geo_rad_km, s_src=s_src,
            s_area=s_area, s_ven=s_ven, s_eve=s_eve, s_class=s_class,
            s_city=s_city, event_token_list=event_token_list,
            request_source_info=request_source_info,
            request_extra_info=request_extra_info,
            request_video_iframe=request_video_iframe,
            request_cost_range=request_cost_range,
            request_media=request_media,
            request_custom_fields=request_custom_fields,
            request_avail_details=request_avail_details,
            s_top=s_top, s_user_rating=s_user_rating,
            s_critic_rating=s_critic_rating,
            s_auto_range=s_auto_range, page_length=page_length,
            page_number=page_number,
            s_cust_fltr=s_cust_fltr,
            request_reviews=request_reviews,
            s_airport=s_airport,
            mime_text_type=mime_text_type,
        )

        return self.parse_response(parse.event_search_result, resp)

    def extra_info(
            self, crypto_block, event_token, upfront_data_token=None,
            source_info=None, request_media=None,
            mime_text_type=None, request_avail_details=None):
        resp = self.make_core_request(
            'extra_info',
            crypto_block=crypto_block, upfront_data_token=upfront_data_token,
            event_token=event_token, source_info=source_info,
            request_media=request_media, mime_text_type=mime_text_type,
            request_avail_details=request_avail_details,
        )

        return self.parse_response(parse.extra_info_result, resp)

    def date_time_options(
            self, crypto_block, event_token, upfront_data_token=None,
            earliest_date=None, latest_date=None, request_cost_range=None,
            page_length=None, page_number=None):
        resp = self.make_core_request(
            'date_time_options',
            crypto_block=crypto_block, event_token=event_token,
            upfront_data_token=upfront_data_token, earliest_date=earliest_date,
            latest_date=latest_date, request_cost_range=request_cost_range,
            page_length=page_length, page_number=page_number,
        )

        return self.parse_response(parse.date_time_options_result, resp)

    def month_options(
            self, crypto_block, event_token, upfront_data_token=None):
        resp = self.make_core_request(
            'month_options',
            crypto_block=crypto_block,
            event_token=event_token,
            upfront_data_token=upfront_data_token,
        )

        return self.parse_response(parse.month_options_result, resp)

    def availability_options(
            self, crypto_block, upfront_data_token=None, perf_token=None,
            departure_date=None, usage_date=None, self_print_mode=None,
            trolley_token=None, add_discounts=None, quantity_options_only=None,
            no_of_tickets=None, add_free_seat_blocks=None,
            add_user_commission=None):

        resp = self.make_core_request(
            'availability_options',
            crypto_block=crypto_block, upfront_data_token=upfront_data_token,
            perf_token=perf_token, departure_date=departure_date,
            usage_date=usage_date, self_print_mode=self_print_mode,
            trolley_token=trolley_token, add_discounts=add_discounts,
            quantity_options_only=quantity_options_only,
            no_of_tickets=no_of_tickets,
            add_free_seat_blocks=add_free_seat_blocks,
            add_user_commission=add_user_commission,
        )

        return self.parse_response(parse.availability_options_result, resp)

    def despatch_options(
            self, crypto_block, upfront_data_token=None, perf_token=None,
            departure_date=None, usage_date=None, self_print_mode=None,
            trolley_token=None):

        resp = self.make_core_request(
            'despatch_options',
            crypto_block=crypto_block, upfront_data_token=upfront_data_token,
            perf_token=perf_token, departure_date=departure_date,
            usage_date=usage_date, self_print_mode=self_print_mode,
            trolley_token=trolley_token,
        )

        return self.parse_response(parse.despatch_options_result, resp)

    def discount_options(
            self, crypto_block, band_token, no_of_tickets,
            upfront_data_token=None, despatch_token=None, trolley_token=None,
            seat_block_token=None, seat_block_offset=None,
            add_user_commission=None):
        resp = self.make_core_request(
            'discount_options',
            crypto_block=crypto_block,
            band_token=band_token, despatch_token=despatch_token,
            no_of_tickets=no_of_tickets, upfront_data_token=upfront_data_token,
            trolley_token=trolley_token, seat_block_token=seat_block_token,
            seat_block_offset=seat_block_offset,
            add_user_commission=add_user_commission,
        )

        return self.parse_response(parse.discount_options_result, resp)

    def create_order(
            self, crypto_block, upfront_data_token=None, discount_token=None,
            despatch_token=None):
        resp = self.make_core_request(
            'create_order',
            crypto_block=crypto_block, upfront_data_token=upfront_data_token,
            discount_token=discount_token, despatch_token=despatch_token,
        )

        return self.parse_response(parse.create_order_result, resp)

    def create_order_and_reserve(
            self, crypto_block, upfront_data_token=None, discount_token=None,
            despatch_token=None):
        resp = self.make_core_request(
            'create_order_and_reserve',
            crypto_block=crypto_block, upfront_data_token=upfront_data_token,
            discount_token=discount_token, despatch_token=despatch_token,
        )

        return self.parse_response(parse.create_order_and_reserve_result, resp)

    def trolley_add_order(
            self, crypto_block, order_token, upfront_data_token=None,
            trolley_token=None, describe_trolley=None):
        resp = self.make_core_request(
            'trolley_add_order',
            crypto_block=crypto_block, order_token=order_token,
            upfront_data_token=upfront_data_token, trolley_token=trolley_token,
            describe_trolley=describe_trolley,
        )

        return self.parse_response(parse.trolley_add_order_result, resp)

    def trolley_describe(
            self, crypto_block, trolley_token, upfront_data_token=None):
        resp = self.make_core_request(
            'trolley_describe',
            crypto_block=crypto_block, trolley_token=trolley_token,
            upfront_data_token=upfront_data_token,
        )

        return self.parse_response(parse.trolley_describe_result, resp)

    def trolley_remove(
            self, crypto_block, trolley_token, upfront_data_token=None,
            remove_item=None, describe_trolley=None):
        resp = self.make_core_request(
            'trolley_remove',
            crypto_block=crypto_block, trolley_token=trolley_token,
            upfront_data_token=upfront_data_token, remove_item=remove_item,
            describe_trolley=describe_trolley,
        )

        return self.parse_response(parse.trolley_remove_result, resp)

    def make_reservation(
            self, crypto_block, trolley_token, upfront_data_token=None,
            self_print_mode=None, describe_trolley=None):
        resp = self.make_core_request(
            'make_reservation',
            crypto_block=crypto_block, trolley_token=trolley_token,
            upfront_data_token=upfront_data_token,
            self_print_mode=self_print_mode,
            describe_trolley=describe_trolley,
        )

        return self.parse_response(parse.make_reservation_result, resp)

    def get_reservation_link(
            self, crypto_block, trolley_token, upfront_data_token=None):
        resp = self.make_core_request(
            'get_reservation_link',
            crypto_block=crypto_block, trolley_token=trolley_token,
            upfront_data_token=upfront_data_token,
        )

        return self.parse_response(parse.get_reservation_link_result, resp)

    def release_reservation(
            self, crypto_block, upfront_data_token=None):
        resp = self.make_core_request(
            'release_reservation',
            crypto_block=crypto_block, upfront_data_token=upfront_data_token,
        )

        return self.parse_response(parse.release_reservation_result, resp)

    def purchase_reservation_part_one(
            self, crypto_block, customer_data, return_token, return_domain,
            return_path, return_with_https, encryption_key, card_data=None,
            user_can_use_data=None, supplier_can_use_data=None,
            world_can_use_data=None, upfront_data_token=None):
        resp = self.make_core_request(
            'purchase_reservation_part_one',
            crypto_block=crypto_block, customer_data=customer_data,
            return_token=return_token, return_domain=return_domain,
            return_path=return_path, return_with_https=return_with_https,
            encryption_key=encryption_key, card_data=card_data,
            user_can_use_data=user_can_use_data,
            supplier_can_use_data=supplier_can_use_data,
            world_can_use_data=world_can_use_data,
            upfront_data_token=upfront_data_token,
        )

        return self.parse_response(
            parse.purchase_reservation_part_one_result, resp
        )

    def purchase_reservation_part_two(
            self, returning_token, new_return_token, new_return_path, http_referer,
            http_accept, http_user_agent, callback_data, encryption_key,
            crypto_block=None, send_confirmation_email=None, results_url=None,
            upfront_data_token=None):

        if crypto_block is None:
            user_passwd = self.password
        else:
            user_passwd = None

        resp = self.make_core_request(
            'purchase_reservation_part_two',
            user_passwd=user_passwd,
            crypto_block=crypto_block,
            upfront_data_token=upfront_data_token,
            returning_token=returning_token,
            new_return_token=new_return_token,
            new_return_path=new_return_path,
            http_referer=http_referer, http_accept=http_accept,
            http_user_agent=http_user_agent, callback_data=callback_data,
            encryption_key=encryption_key,
            send_confirmation_email=send_confirmation_email,
            results_url=results_url,
        )

        return self.parse_response(
            parse.purchase_reservation_part_two_result, resp
        )

    def purchase_reservation(
            self, crypto_block, customer_data, card_data=None,
            send_confirmation_email=None,
            upfront_data_token=None):

        resp = self.make_core_request(
            'purchase_reservation',
            crypto_block=crypto_block, customer_data=customer_data,
            card_data=card_data,
            send_confirmation_email=send_confirmation_email,
            upfront_data_token=upfront_data_token,
        )

        return self.parse_response(
            parse.purchase_reservation, resp
        )

    def transaction_info(
            self, transaction_id, describe_trolley=None, describe_customer=None,
            describe_external_sale_page=None, crypto_block=None,
            upfront_data_token=None):

        if crypto_block is None:
            user_passwd = self.password
        else:
            user_passwd = None

        resp = self.make_core_request(
            'transaction_info',
            user_passwd=user_passwd,
            transaction_id=transaction_id,
            describe_trolley=describe_trolley,
            describe_customer=describe_customer,
            describe_external_sale_page=describe_external_sale_page,
            crypto_block=crypto_block, upfront_data_token=upfront_data_token,
        )

        return self.parse_response(parse.transaction_info_result, resp)

    def save_external_sale_page(
            self, transaction_id, sale_page_type, sale_page_subtype, sale_page,
            crypto_block=None, upfront_data_token=None):

        if crypto_block is None:
            user_passwd = self.password
        else:
            user_passwd = None

        resp = self.make_core_request(
            'save_external_sale_page',
            user_passwd=user_passwd,
            crypto_block=crypto_block,
            upfront_data_token=upfront_data_token,
            transaction_id=transaction_id,
            sale_page_type=sale_page_type,
            sale_page_subtype=sale_page_subtype,
            sale_page=sale_page,
        )

        return self.parse_response(parse.save_external_sale_page_result, resp)



import unittest
import datetime
from copy import deepcopy

from pyticketswitch.interface_objects import (
    Core, Event, Concession, DespatchMethod, TicketType, Performance,
    Trolley, Order, Reservation, Review, Seat, Bundle, Address, Customer,
)
from pyticketswitch.api_exceptions import InvalidId
from pyticketswitch import settings_test as settings


class InterfaceObjectTestCase(unittest.TestCase):
    api_settings = {
        'username': settings.TEST_USERNAME,
        'password': settings.TEST_PASSWORD,
        'url': settings.API_URL,
        'ext_start_session_url': settings.EXT_START_SESSION_URL
    }


class CoreTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        self.core = Core(session=session, **self.api_settings)

    def test_search_events_keyword(self):

        events = self.core.search_events(keyword='test')

        self.assertTrue(events)

    def test_search_events_contains_event(self):

        events = self.core.search_events(keyword='test')

        for e in events:
            self.assertIsInstance(e, Event)

    def test_search_events_keyword_paging(self):

        events = self.core.search_events(
            keyword='test', page_length='2', page_number='1'
        )

        self.assertTrue(events)

    def test_search_events_keyword_empty(self):

        events = self.core.search_events(keyword='abc24382hggeouh%$')

        self.assertFalse(events)

    def test_search_events_sort_top(self):

        events = self.core.search_events(
            keyword='test', sort_by='top'
        )

        self.assertTrue(events)

    def test_search_events_sort_user_rating(self):

        events = self.core.search_events(
            keyword='test', sort_by='user_rating'
        )

        self.assertTrue(events)

    def test_search_events_sort_critic_rating(self):

        events = self.core.search_events(
            keyword='test', sort_by='critic_rating'
        )

        self.assertTrue(events)

    def test_search_events_auto_date_range(self):

        for value in settings.AUTO_DATE_RANGE:
            events = self.core.search_events(
                keyword='test', auto_date_range=value
            )

            self.assertTrue(events)

    def test_search_events_cities(self):

        self.core.search_events(keyword='test')

        self.assertIsInstance(self.core.event_cities, dict)

        self.assertIn('london-uk', self.core.event_cities)

        for city, info in self.core.event_cities.items():
            self.assertIsInstance(city, str)
            self.assertIsInstance(info, dict)

            self.assertIn('count', info)
            self.assertIn('description', info)

    def test_search_events_categories(self):

        self.core.search_events(keyword='test')

        self.assertIsInstance(self.core.event_categories, dict)

        self.assertIn('theatre/', self.core.event_categories)

        for cat, info in self.core.event_categories.items():
            self.assertIsInstance(cat, str)
            self.assertIsInstance(info, dict)

            self.assertIn('count', info)
            self.assertIn('category', info)
            self.assertIn('sub_categories', info)

    def test_search_events_countries(self):

        self.core.search_events(keyword='test')

        self.assertIsInstance(self.core.event_countries, dict)

        self.assertIn('uk', self.core.event_countries)

        for country, info in self.core.event_countries.items():
            self.assertIsInstance(country, str)
            self.assertIsInstance(info, dict)

            self.assertIn('count', info)
            self.assertIn('description', info)

    def test_search_events_price_range(self):

        self.core.search_events(keyword='test')

        self.assertTrue(self.core.event_price_range)

        self.assertIsInstance(self.core.event_price_range, list)

        for price in self.core.event_price_range:
            self.assertIsInstance(price, float)


class InvalidEventTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        self.event = Event(
            event_id='invalid', session=session, **self.api_settings
        )

    def test_get_details(self):
        self.assertRaises(InvalidId, self.event.get_details)

    def test_get_performances(self):
        self.assertRaises(InvalidId, self.event.get_performances)

    def test_get_valid_months(self):
        self.assertRaises(InvalidId, self.event.get_valid_months)

    def test_property(self):
        self.assertFalse(self.event.description)


class ValidEventTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        self.event = Event(
            event_id='6IF', session=session, **self.api_settings)

    def test_string_properties(self):

        for prop_name in (
            'description', 'information', 'venue_desc',
            'venue_info', 'venue_addr', 'supplier_desc',
            'city_code', 'city_desc', 'country_code',
            'country_desc', 'latitude', 'longitude',
            'event_id',
        ):
            self.assertIsInstance(getattr(self.event, prop_name), str)

    def test_categories(self):

        found = False

        for c in self.event.categories:
            if c.code == 'theatre':
                found = True

        self.assertTrue(
            found,
            msg="'Theatre' category not found in event category list"
        )

    def test_float_properties(self):

        for prop_name in (
            'min_seatprice_float',
        ):
            self.assertIsInstance(getattr(self.event, prop_name), float)

    def test_unicode_properties(self):

        for prop_name in (
            'min_seatprice', 'min_combined_price',
        ):
            self.assertIsInstance(getattr(self.event, prop_name), unicode)

    def test_boolean_properties(self):

        for prop_name in (
            'has_no_perfs',
        ):
            self.assertIsInstance(getattr(self.event, prop_name), bool)

    def test_images(self):

        for img in (
            # 'get_large_image',
            'get_medium_image',
            'get_small_image',
            # 'get_seating_plan',
        ):
            self.assertIsNotNone(getattr(self.event, img))

    def test_performances_is_list(self):
        self.assertIsInstance(self.event.get_performances(), list)

    def test_all_performances(self):
        self.assertTrue(self.event.get_performances())

    def test_performances_contains_performance(self):

        performances = self.event.get_performances()

        for perf in performances:
            self.assertIsInstance(perf, Performance)

    def test_no_performances_old_date(self):

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=-1)
        )

        self.assertFalse(
            self.event.get_performances(
                latest_date=latest_date
            )
        )

    def test_performances_one_week(self):

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=6)
        )

        self.assertTrue(
            self.event.get_performances(
                latest_date=latest_date
            )
        )

    def test_performance_calendar(self):
        self.event.get_performances()

        self.assertTrue(self.event.performance_calendar)

    def test_performance_calendar_empty(self):

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=-1)
        )

        self.event.get_performances(latest_date=latest_date)

        self.assertFalse(self.event.performance_calendar)


class EventReviewsTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        self.event = Event(
            event_id='6IF', session=session, **self.api_settings)

    def test_critic_review_percent(self):
        self.assertIsInstance(self.event.critic_review_percent, str)

    def test_critic_reviews(self):
        self.assertTrue(self.event.critic_reviews)

    def test_reviews_contains_review(self):
        for r in self.event.critic_reviews:
            self.assertIsInstance(r, Review)


class ReviewTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        self.review = event.critic_reviews[0]

    def test_review_is_user_review(self):
        self.assertFalse(self.review.is_user_review)

    def test_review_string_properties(self):

        for prop_name in (
            'title', 'author', 'date_desc', 'time_desc', 'body',
        ):
            self.assertIsInstance(getattr(self.review, prop_name), str)

    def test_date_is_datetime_date(self):
        self.assertIsInstance(self.review.date, datetime.date)

    def test_time_is_datetime_time(self):
        self.assertIsInstance(self.review.time, datetime.time)


class AvailDetailTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        self.avail_detail = event.avail_details[0]

    def test_string_properties(self):

        for prop_name in (
            'ticket_type_code', 'ticket_type_desc',
            'price_band_code', 'price_band_desc',
        ):
            prop = getattr(self.avail_detail, prop_name)
            if prop is not None:
                self.assertIsInstance(prop, str)

    def test_unicode_properties(self):

        for prop_name in (
            'seatprice', 'surcharge', 'price_combined', 'non_offer_seatprice',
            'non_offer_surcharge', 'non_offer_combined', 'absolute_saving',
            'percentage_saving',
        ):
            prop = getattr(self.avail_detail, prop_name)
            if prop is not None:
                self.assertIsInstance(prop, unicode)

    def test_weekday_available(self):
        self.assertIsInstance(self.avail_detail.weekdays_available, list)
        self.assertEqual(len(self.avail_detail.weekdays_available), 7)
        for day in self.avail_detail.weekdays_available:
            self.assertIsInstance(day, bool)

    def test_date_properties(self):

        for prop_name in (
            'available_from_date', 'available_until_date',
        ):
            prop = getattr(self.avail_detail, prop_name)
            if prop is not None:
                self.assertIsInstance(prop, datetime.date)

    def test_float_properties(self):

        for prop_name in (
            'price_combined_float', 'non_offer_combined_float',
            'seatprice_float', 'non_offer_seatprice_float',
            'surcharge_float', 'non_offer_surcharge_float',
            'absolute_saving_float',
        ):
            prop = getattr(self.avail_detail, prop_name)
            if prop is not None:
                self.assertIsInstance(prop, float)

    def test_int_properties(self):

        for prop_name in (
            'int_percentage_saving',
        ):
            prop = getattr(self.avail_detail, prop_name)
            if prop is not None:
                self.assertIsInstance(prop, int)

    def test_bool_properties(self):

        for prop_name in (
            'has_no_booking_fee',
        ):
            prop = getattr(self.avail_detail, prop_name)
            self.assertIsInstance(prop, bool)

    def test_comparison(self):

        avail_detail_dup = deepcopy(self.avail_detail)

        self.assertTrue(
            self.avail_detail.is_same_ticket_and_price(avail_detail_dup)
        )
        avail_detail_dup._ticket_type_desc = 'Test'
        self.assertFalse(
            self.avail_detail.is_same_ticket_and_price(avail_detail_dup)
        )


class PerformanceTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        self.performances = event.get_performances(
            latest_date=latest_date
        )

        self.performance = self.performances[0]

    def test_string_properties(self):

        for prop_name in (
            'date_desc', 'time_desc', 'perf_id'
        ):
            self.assertIsInstance(getattr(self.performance, prop_name), str)

    def test_date_is_datetime_date(self):
        self.assertIsInstance(self.performance.date, datetime.date)

    def test_time_is_datetime_time(self):
        self.assertIsInstance(self.performance.time, datetime.time)

    def test_required_info(self):

        for perf in self.performances:
            if perf.date.weekday() == 1:
                self.assertIsInstance(perf.required_info, str)


class PerformanceAvailabilityTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                self.performance = p
                break

        self.performance.get_availability()

    def test_ticket_types(self):
        self.assertTrue(self.performance.ticket_types)

    def test_ticket_types_contains_ticket_type(self):
        for tt in self.performance.ticket_types:
            self.assertIsInstance(tt, TicketType)

    def test_despatch_methods(self):
        self.assertTrue(self.performance.despatch_methods)

    def test_despatch_methods_contains_despatch_method(self):
        for dm in self.performance.despatch_methods:
            self.assertIsInstance(dm, DespatchMethod)

    def test_valid_ticket_quantities(self):
        self.assertTrue(self.performance.valid_ticket_quantities)

    def test_valid_ticket_quantities_contains_int(self):
        for quantity in self.performance.valid_ticket_quantities:
            self.assertIsInstance(quantity, int)

    def test_availability_returned_event(self):
        self.assertTrue(self.performance.event)

    def test_list_properties(self):
        for prop_name in (
            'ticket_types', 'despatch_methods', 'valid_ticket_quantities',
            'ticket_types_by_combined_price', 'ticket_types_by_saving',
            'unique_combined_prices'
        ):
            self.assertIsInstance(getattr(self.performance, prop_name), list)

    def test_sorted_combined_price(self):
        prev_combined_price = 0
        for tt in self.performance.ticket_types_by_combined_price:
            self.assertGreaterEqual(
                tt.price_combined_float, prev_combined_price
            )

    def test_unique_combined_price(self):
        values = []
        for unique_value in self.performance.unique_combined_prices:
            self.assertNotIn(unique_value, values)
            values.append(unique_value)

# Need tests for usage date/departure date somewhere


class PerformanceDespatchTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        self.performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(self.performances):
            if p.date.weekday() != 5:
                self.performance = p
                break

    def test_despatch_methods(self):
        self.assertTrue(self.performance.get_despatch_methods())

    def test_despatch_methods_contains_despatch_method(self):
        for dm in self.performance.get_despatch_methods():
            self.assertIsInstance(dm, DespatchMethod)

    def test_despatch_methods_is_list(self):
        self.assertIsInstance(self.performance.get_despatch_methods(), list)

    def test_despatch_methods_multi_returned(self):
        self.assertGreater(len(self.performance.get_despatch_methods()), 1)

    def test_despatch_returned_event(self):
        self.performance.get_despatch_methods()

        self.assertTrue(self.performance.event)


class TicketTypeTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability(
            include_user_commission=True
        )

        self.ticket_type = ticket_types[0]

    def test_quantity_options(self):
        self.assertIsInstance(self.ticket_type.quantity_options, list)

        for option in self.ticket_type.quantity_options:
            self.assertIsInstance(option, tuple)

            for item in option:
                self.assertIsInstance(item, str)

    def test_string_properties(self):

        for prop_name in (
            'description', 'ticket_type_id',
        ):
            self.assertIsInstance(getattr(self.ticket_type, prop_name), str)

    def test_float_properties(self):

        for prop_name in (
            'non_offer_combined_float', 'price_combined_float',
            'surcharge_float', 'price_without_surcharge_float',
        ):
            self.assertIsInstance(getattr(self.ticket_type, prop_name), float)

    def test_unicode_properties(self):

        for prop_name in (
            'price_without_surcharge', 'surcharge', 'price_combined',
            'non_offer_combined'
        ):
            self.assertIsInstance(
                getattr(self.ticket_type, prop_name), unicode
            )

    def test_int_properties(self):

        for prop_name in (
            'int_percentage_saving', 'number_available',
        ):
            self.assertIsInstance(getattr(self.ticket_type, prop_name), int)

    def test_example_seats(self):
        self.assertIsInstance(self.ticket_type.example_seats, list)

        for ex in self.ticket_type.example_seats:
            self.assertIsInstance(ex, Seat)

    # def test_user_commission(self):
    #     user_commission = self.ticket_type.user_commission
    #     self.assertIsInstance(user_commission, Commission)
    #     self.assertIsInstance(user_commission.amount_excluding_vat, float)
    #     self.assertIsInstance(user_commission.amount_including_vat, float)
    #     self.assertIsInstance(user_commission.commission_currency, Currency)


class TicketTypeConcessionTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability()

        self.ticket_type = ticket_types[0]

        self.ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=performance.despatch_methods[0]
        )

    def test_blanket_discount_only(self):
        self.assertIsInstance(self.ticket_type.blanket_discount_only, str)

    def test_ticket_concessions(self):
        self.assertIsInstance(self.ticket_type.ticket_concessions, list)

        for concessions in self.ticket_type.ticket_concessions:
            self.assertIsInstance(concessions, list)

            for concession in concessions:
                self.assertIsInstance(concession, Concession)


class ConcessionTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability()

        ticket_type = ticket_types[0]

        ticket_concessions = ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=performance.despatch_methods[0]
        )

        concessions = ticket_concessions[0]

        self.concession = concessions[0]

    def test_string_properties(self):

        for prop_name in (
            'concession_id', 'description',
        ):
            self.assertIsInstance(getattr(self.concession, prop_name), str)

    def test_unicode_properties(self):

        for prop_name in (
            'ticket_price', 'surcharge', 'seatprice'
        ):
            self.assertIsInstance(getattr(self.concession, prop_name), unicode)

    def test_float_properties(self):

        for prop_name in (
            'ticket_price_float', 'surcharge_float', 'seatprice_float'
        ):
            self.assertIsInstance(getattr(self.concession, prop_name), float)

    # def test_seat_codes(self):

    #     self.assertIsInstance(self.concession.seat_codes, list)

    #     for seat_id in self.concession.seat_codes:
    #         self.assertIsInstance(seat_id, str)


class DespatchMethodTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        despatch_methods = performance.get_despatch_methods()

        self.despatch_method = despatch_methods[0]

    def test_string_properties(self):

        for prop_name in (
            'description', 'despatch_id',
        ):
            self.assertIsInstance(
                getattr(self.despatch_method, prop_name), str
            )

    def test_unicode_properties(self):

        for prop_name in (
            'cost',
        ):
            self.assertIsInstance(
                getattr(self.despatch_method, prop_name), unicode
            )


class OrderTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability()

        ticket_type = ticket_types[0]

        despatch_method = performance.despatch_methods[0]

        ticket_concessions = ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=despatch_method
        )

        concessions = ticket_concessions[0]

        concession = concessions[0]

        concession_list = [concession]

        core = Core(session=session, **self.api_settings)

        self.order = core.create_order(
            concessions=concession_list,
            despatch_method=despatch_method,
        )

    def test_string_properties(self):

        for prop_name in (
            'event_desc', 'venue_desc',
            'ticket_type_desc', 'order_id',
        ):
            self.assertIsInstance(
                getattr(self.order, prop_name), str
            )

    def test_int_properties(self):

        for prop_name in (
            'item_number',
        ):
            self.assertIsInstance(
                getattr(self.order, prop_name), int
            )

    def test_unicode_properties(self):

        for prop_name in (
            'total_combined',
        ):
            self.assertIsInstance(
                getattr(self.order, prop_name), unicode
            )

    def test_performance(self):
        self.assertIsInstance(self.order.performance, Performance)

    def test_event(self):
        self.assertIsInstance(self.order.event, Event)

    def test_concession(self):
        self.assertIsInstance(self.order.concessions[0], Concession)

    def test_despatch_method(self):
        self.assertIsInstance(self.order.despatch_method, DespatchMethod)


class TrolleyTests(InterfaceObjectTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability()

        ticket_type = ticket_types[0]

        despatch_method = performance.despatch_methods[0]

        ticket_concessions = ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=despatch_method
        )

        concessions = ticket_concessions[0]

        concession = concessions[0]

        concession_list = [concession]

        core = Core(session=session, **self.api_settings)

        order = core.create_order(
            concessions=concession_list,
            despatch_method=despatch_method,
        )

        self.trolley = Trolley(session=session, **self.api_settings)

        self.trolley.add_order(order)

    def test_string_properties(self):

        for prop_name in (
            'trolley_id',
        ):
            self.assertIsInstance(
                getattr(self.trolley, prop_name), str
            )

    def test_int_properties(self):

        for prop_name in (
            'order_count',
        ):
            self.assertIsInstance(
                getattr(self.trolley, prop_name), int
            )

    def test_update(self):
        session = {}
        new_trolley = Trolley(
            trolley_id=self.trolley.trolley_id,
            session=session,
            **self.api_settings
        )

        new_trolley.update()

        self.assertTrue(new_trolley.order_count)

    def test_bundles(self):

        self.assertIsInstance(self.trolley.bundles, list)

        for o in self.trolley.bundles:
            self.assertIsInstance(o, Bundle)

    def test_orders(self):

        self.assertIsInstance(self.trolley.orders, list)

        for o in self.trolley.orders:
            self.assertIsInstance(o, Order)


class BundleTests(InterfaceObjectTestCase):

    def get_order(self, event_id):
        session = {}
        event = Event(event_id=event_id, session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_type = performance.get_availability()[0]

        despatch_method = performance.despatch_methods[0]

        concession = ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=despatch_method
        )[0][0]

        concession_list = [concession]

        core = Core(session=session, **self.api_settings)

        return core.create_order(
            concessions=concession_list,
            despatch_method=despatch_method,
        )

    def setUp(self):

        order1 = self.get_order('6IF')
        order2 = self.get_order('6L9')

        session = {}
        self.trolley = Trolley(session=session, **self.api_settings)

        self.trolley.add_order(order1)
        self.trolley.add_order(order2)

    def test_string_properties(self):

        for prop_name in (
            'source_description', 'source_code',
        ):
            self.assertIsInstance(
                getattr(self.trolley.bundles[0], prop_name), str
            )

    def test_unicode_properties(self):

        for prop_name in (
            'total_seatprice', 'total_surcharge',
            'total_despatch', 'total_cost',
        ):
            self.assertIsInstance(
                getattr(self.trolley.bundles[0], prop_name), unicode
            )

    def test_int_properties(self):

        for prop_name in (
            'order_count',
        ):
            self.assertIsInstance(
                getattr(self.trolley.bundles[0], prop_name), int
            )

    def test_float_properties(self):

        for prop_name in (
            'total_cost_float',
        ):
            self.assertIsInstance(
                getattr(self.trolley.bundles[0], prop_name), float
            )

    def test_bundles_count(self):

        self.assertEqual(
            len(self.trolley.bundles),
            2,
        )


class ReservationTests(object):

    def test_reservation_is_reservation(self):
        self.assertIsInstance(self.reservation, Reservation)

    def test_string_properties(self):

        for prop_name in (
            'transaction_id',
        ):
            self.assertIsInstance(
                getattr(self.reservation, prop_name), str
            )

    def test_bool_properties(self):

        for prop_name in (
            'need_payment_card',
            'needs_email_address', 'supports_billing_address',
            'needs_agent_reference',

        ):
            self.assertIsInstance(
                getattr(self.reservation, prop_name), bool
            )

    def test_get_details(self):
        session = {}
        new_reservation = Reservation(
            transaction_id=self.reservation.transaction_id,
            session=session,
            **self.api_settings
        )

        trolley = new_reservation.get_details()

        self.assertIsInstance(trolley, Trolley)


class MakeReservationTests(InterfaceObjectTestCase, ReservationTests):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability()

        ticket_type = ticket_types[0]

        despatch_method = performance.despatch_methods[0]

        ticket_concessions = ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=despatch_method
        )

        concessions = ticket_concessions[0]

        concession = concessions[0]

        concession_list = [concession]

        core = Core(session=session, **self.api_settings)

        order = core.create_order(
            concessions=concession_list,
            despatch_method=despatch_method,
        )

        self.trolley = Trolley(session=session, **self.api_settings)

        self.trolley.add_order(order)

        self.reservation = self.trolley.get_reservation()


class CombinedOrderReservationTests(InterfaceObjectTestCase, ReservationTests):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability()

        ticket_type = ticket_types[0]

        despatch_method = performance.despatch_methods[0]

        ticket_concessions = ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=despatch_method
        )

        concessions = ticket_concessions[0]

        concession = concessions[0]

        concession_list = [concession]

        core = Core(session=session, **self.api_settings)

        self.reservation = core.create_reservation(
            concessions=concession_list,
            despatch_method=despatch_method,
        )


class InterfaceObjectCreditUserTestCase(unittest.TestCase):
    api_settings = {
        'username': settings.TEST_CREDIT_USERNAME,
        'password': settings.TEST_CREDIT_PASSWORD,
        'url': settings.API_URL,
        'ext_start_session_url': settings.EXT_START_SESSION_URL
    }


class PurchaseReservationOnCreditTests(InterfaceObjectCreditUserTestCase):

    def setUp(self):
        session = {}
        event = Event(event_id='6IF', session=session, **self.api_settings)

        latest_date = (
            datetime.date.today() +
            datetime.timedelta(days=7)
        )

        performances = event.get_performances(
            latest_date=latest_date
        )

        for p in reversed(performances):
            if p.date.weekday() != 5:
                performance = p
                break

        ticket_types = performance.get_availability()

        ticket_type = ticket_types[0]

        despatch_method = performance.despatch_methods[0]

        ticket_concessions = ticket_type.get_concessions(
            no_of_tickets=1,
            despatch_method=despatch_method
        )

        concessions = ticket_concessions[0]

        concession = concessions[0]

        concession_list = [concession]

        core = Core(session=session, **self.api_settings)

        order = core.create_order(
            concessions=concession_list,
            despatch_method=despatch_method,
        )

        self.trolley = Trolley(session=session, **self.api_settings)

        self.trolley.add_order(order)

        self.reservation = self.trolley.get_reservation()

        customer_address = Address(
            address_line_one='1 Test Lane',
            address_line_two=None,
            town='Test Town',
            county='County',
            postcode='AB12 3CD',
            country_code='GB'
        )

        customer = Customer(
            title='Mr',
            first_name='Test',
            last_name='User',
            email_address='test@test.com',
            home_phone='01234567890',
            work_phone='01234567890',
            address=customer_address,
            user_can_use_data=True,
            supplier_can_use_data=False,
            world_can_use_data=False
        )

        self.purchase_response = self.reservation.purchase_reservation(
            customer=customer,
        )

    def test_purchase_succeeded(self):
        self.assertTrue(self.reservation.is_purchased)
